import functools
import logging
from typing import Any, Dict, List, Optional, Tuple

from mongo_analyser.core import db as core_db_manager
from mongo_analyser.core.shared import redact_uri_password
from mongo_analyser.dialogs import ErrorDialog
from pymongo.errors import ConnectionFailure as PyMongoConnectionFailure
from pymongo.errors import OperationFailure as PyMongoOperationFailure
from rich.text import Text
from textual import on
from textual.app import ComposeResult
from textual.containers import Container
from textual.css.query import NoMatches
from textual.reactive import reactive
from textual.widgets import Button, DataTable, Input, Label, Static
from textual.worker import Worker, WorkerCancelled

logger = logging.getLogger(__name__)


def _is_auth_error(e: PyMongoOperationFailure) -> bool:
    """Checks if a PyMongoOperationFailure is likely an authorization error."""
    if e.code == 13:
        return True
    error_msg_lower = str(e).lower()
    if "not authorized" in error_msg_lower or "unauthorized" in error_msg_lower:
        return True
    return False


class DBConnectionView(Container):
    connection_status = reactive(Text.from_markup("[#D08770]Not Connected[/]"))

    def compose(self) -> ComposeResult:
        yield Label("MongoDB URI:")
        yield Input(id="mongo_uri_input", value="mongodb://localhost:27017/")
        yield Label("Database Name (optional, overrides URI's DB if specified):")
        yield Input(id="mongo_db_name_input", placeholder="my_database")
        yield Button("Connect to DB", variant="primary", id="connect_mongo_button")
        yield Static(self.connection_status, id="mongo_connection_status_label")

        yield Label(
            "Collections in Database (Click to Select):",
            classes="panel_title_small",
            id="collections_title_label",
        )
        yield DataTable(
            id="collections_data_table",
            show_header=True,
            show_cursor=True,
            cursor_type="row",
            classes="collections_list_container",
        )

        yield Label(
            "Indexes for Selected Collection:",
            classes="panel_title_small",
            id="indexes_title_label",
        )
        yield DataTable(
            id="indexes_data_table",
            show_header=True,
            show_cursor=False,
        )

    def on_mount(self) -> None:
        collections_table = self.query_one("#collections_data_table", DataTable)
        if not collections_table.columns:
            collections_table.add_columns(
                "Name", "Docs", "Avg Size", "Total Size", "Storage Size", "Indexes"
            )
        collections_table.visible = False
        self.query_one("#collections_title_label", Label).visible = False

        indexes_table = self.query_one("#indexes_data_table", DataTable)
        if not indexes_table.columns:
            indexes_table.add_columns(
                "Name", "Fields (Key)", "Unique", "Sparse", "Background", "Other Props"
            )
        indexes_table.visible = False
        self.query_one("#indexes_title_label", Label).visible = False

        try:
            uri_input_widget = self.query_one("#mongo_uri_input", Input)
            db_name_input_widget = self.query_one("#mongo_db_name_input", Input)

            initial_uri_from_app = getattr(self.app, "_initial_mongo_uri", None)
            initial_db_name_from_app = getattr(self.app, "_initial_db_name", None)

            if initial_uri_from_app:
                uri_input_widget.value = initial_uri_from_app
                if logger.isEnabledFor(logging.INFO):
                    logger.info(
                        "DBConnectionView pre-filled Mongo URI from initial config: %s",
                        redact_uri_password(initial_uri_from_app),
                    )

            if initial_db_name_from_app:
                db_name_input_widget.value = initial_db_name_from_app
                if logger.isEnabledFor(logging.INFO):
                    logger.info(
                        "DBConnectionView pre-filled Mongo Database Name from initial config: %s",
                        initial_db_name_from_app,
                    )

        except NoMatches:
            if logger.isEnabledFor(logging.WARNING):
                logger.warning(
                    "DBConnectionView: URI or DB Name input widgets not found on mount for pre-filling."
                )
        except Exception as e:
            if logger.isEnabledFor(logging.ERROR):
                logger.error(
                    "Error pre-filling DBConnectionView inputs from initial app config: %s",
                    e,
                    exc_info=True,
                )

        self.focus_default_widget()

    def focus_default_widget(self) -> None:
        try:
            self.query_one("#mongo_uri_input", Input).focus()
        except NoMatches:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("DBConnectionView: Could not focus default input '#mongo_uri_input'.")

    def watch_connection_status(self, new_status: Text) -> None:
        if self.is_mounted:
            try:
                self.query_one("#mongo_connection_status_label", Static).update(new_status)
            except NoMatches:
                pass

    async def _connect_and_list_collections_task(
        self, uri: str, db_name_from_input: Optional[str]
    ) -> Tuple[bool, Text, List[Dict[str, Any]], Optional[str], Optional[str]]:
        collections_with_stats: List[Dict[str, Any]] = []
        final_status_message: Text
        connection_is_meaningful_for_analyser = False
        redacted_uri_for_log = redact_uri_password(uri)
        self._last_connect_error_msg = ""

        if not core_db_manager.db_connection_active(
            uri=uri, db_name=db_name_from_input, server_timeout_ms=3000, force_reconnect=True
        ):
            if "Cannot determine database context" in getattr(
                self, "_internal_db_conn_error_reason", ""
            ):
                final_status_message = Text.from_markup(
                    f"[#BF616A]Connection Failed: MongoDB URI ('{redacted_uri_for_log}') does not specify a default database path, "
                    "and no database name was provided. Please specify a database."
                )
                self._last_connect_error_msg = final_status_message.plain
            else:
                final_status_message = Text.from_markup(
                    f"[#BF616A]Connection Failed: Could not connect to MongoDB server at {redacted_uri_for_log}[/]"
                )
                self._last_connect_error_msg = final_status_message.plain
            return False, final_status_message, [], None, None

        client = core_db_manager.get_mongo_client()
        db_instance = core_db_manager.get_mongo_db()

        if client is None or db_instance is None:
            final_status_message = Text.from_markup(
                "[#BF616A]Internal Error: Could not retrieve active MongoDB client/db after connection attempt.[/]"
            )
            self._last_connect_error_msg = final_status_message.plain
            return (
                False,
                final_status_message,
                [],
                core_db_manager.get_current_uri(),
                core_db_manager.get_current_resolved_db_name(),
            )

        actual_db_name = db_instance.name
        connected_uri = core_db_manager.get_current_uri()

        try:
            collection_names_list = sorted(db_instance.list_collection_names())
            connection_is_meaningful_for_analyser = True

            if collection_names_list:
                for name in collection_names_list:
                    coll_stat_entry = {
                        "name": name,
                        "count": "N/A",
                        "avgObjSize": "N/A",
                        "size": "N/A",
                        "storageSize": "N/A",
                        "nindexes": "N/A",
                    }
                    try:
                        coll_stats = db_instance.command("collStats", name)
                        coll_stat_entry.update(
                            {
                                "count": coll_stats.get("count", "N/A"),
                                "avgObjSize": _format_bytes_tui(coll_stats.get("avgObjSize")),
                                "size": _format_bytes_tui(coll_stats.get("size")),
                                "storageSize": _format_bytes_tui(coll_stats.get("storageSize")),
                                "nindexes": str(coll_stats.get("nindexes", "N/A")),
                            }
                        )
                    except PyMongoOperationFailure as e_stats:
                        logger.warning(
                            "Could not get stats for collection '%s' in DB '%s': %s",
                            name,
                            actual_db_name,
                            e_stats,
                        )
                        if _is_auth_error(e_stats):
                            coll_stat_entry.update(
                                {
                                    k: "Unauthorized"
                                    for k in [
                                        "count",
                                        "avgObjSize",
                                        "size",
                                        "storageSize",
                                        "nindexes",
                                    ]
                                }
                            )

                        else:
                            coll_stat_entry.update(
                                {
                                    k: "Err(Stats)"
                                    for k in [
                                        "count",
                                        "avgObjSize",
                                        "size",
                                        "storageSize",
                                        "nindexes",
                                    ]
                                }
                            )
                    except Exception as e_stats_other:
                        logger.warning(
                            "Unexpected error getting stats for collection '%s' in DB '%s': %s",
                            name,
                            actual_db_name,
                            e_stats_other,
                            exc_info=True,
                        )
                        coll_stat_entry.update(
                            {
                                k: "Err(Stats)"
                                for k in ["count", "avgObjSize", "size", "storageSize", "nindexes"]
                            }
                        )
                    collections_with_stats.append(coll_stat_entry)
                final_status_message = Text.from_markup(
                    f"[#A3BE8C]Connected to {redact_uri_password(connected_uri or 'unknown URI')} (DB: {actual_db_name}). "
                    f"{len(collection_names_list)} collection(s) found.[/]"
                )
            else:
                final_status_message = Text.from_markup(
                    f"[#A3BE8C]Connected to DB: '{actual_db_name}' at {redact_uri_password(connected_uri or 'unknown URI')}. "
                    f"This database is empty (no collections).[/]"
                )
        except PyMongoOperationFailure as e_list_coll:
            connection_is_meaningful_for_analyser = False

            logger.error(
                "MongoDB operation failure listing collections for DB '%s': %s",
                actual_db_name,
                e_list_coll,
                exc_info=not _is_auth_error(e_list_coll),
            )
            if _is_auth_error(e_list_coll):
                final_status_message = Text.from_markup(
                    f"[#BF616A]Connected to server ({redact_uri_password(connected_uri or 'unknown URI')}). However, user is not authorized "
                    f"to list collections in database '{actual_db_name}'. Please check permissions.[/]"
                )
            else:
                err_details = (
                    e_list_coll.details.get("errmsg", str(e_list_coll))
                    if e_list_coll.details
                    else str(e_list_coll)
                )
                final_status_message = Text.from_markup(
                    f"[#BF616A]Connected ({redact_uri_password(connected_uri or 'unknown URI')}, DB: {actual_db_name}), but an error occurred "
                    f"listing collections: {str(err_details)[:70]}[/]"
                )
        except Exception as e_generic_list_coll:
            connection_is_meaningful_for_analyser = False
            logger.error(
                "Unexpected error listing collections for DB '%s': %s",
                actual_db_name,
                e_generic_list_coll,
                exc_info=True,
            )
            final_status_message = Text.from_markup(
                f"[#BF616A]Connected ({redact_uri_password(connected_uri or 'unknown URI')}, DB: {actual_db_name}), but an unexpected error "
                f"occurred while listing collections: {str(e_generic_list_coll)[:70]}[/]"
            )

        if not connection_is_meaningful_for_analyser:
            self._last_connect_error_msg = (
                final_status_message.plain if final_status_message else "Unknown connection error"
            )

        return (
            connection_is_meaningful_for_analyser,
            final_status_message,
            collections_with_stats,
            connected_uri,
            actual_db_name,
        )

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "connect_mongo_button":
            collections_table = self.query_one("#collections_data_table", DataTable)
            collections_title = self.query_one("#collections_title_label", Label)
            indexes_table = self.query_one("#indexes_data_table", DataTable)
            indexes_title = self.query_one("#indexes_title_label", Label)

            uri_input_value = self.query_one("#mongo_uri_input", Input).value
            db_name_input_value = (
                self.query_one("#mongo_db_name_input", Input).value.strip() or None
            )

            self.connection_status = Text.from_markup("[#EBCB8B]Connecting...[/]")
            self.app.available_collections = []
            self.app.active_collection = None

            collections_table.clear()
            indexes_table.clear()
            collections_title.visible = False
            collections_table.visible = False
            indexes_title.visible = False
            indexes_table.visible = False

            try:
                task_with_args = functools.partial(
                    self._connect_and_list_collections_task, uri_input_value, db_name_input_value
                )

                worker_result_type = Tuple[
                    bool, Text, List[Dict[str, Any]], Optional[str], Optional[str]
                ]
                worker: Worker[worker_result_type] = self.app.run_worker(
                    task_with_args,
                    thread=True,
                    name=f"connect_worker_{db_name_input_value or 'from_uri'}",
                    group="db_operations",
                    description=f"Connecting to MongoDB: {redact_uri_password(uri_input_value)}",
                )
                (
                    task_success,
                    status_msg_text,
                    collections_stats_data,
                    connected_uri_result,
                    connected_db_name_actual,
                ) = await worker.wait()

                if worker.is_cancelled:
                    self.connection_status = Text.from_markup(
                        "[#D08770]Connection attempt cancelled.[/]"
                    )
                    self.app.current_mongo_uri = None
                    self.app.current_db_name = None
                    self.app.available_collections = []
                    self.app.active_collection = None
                    return

                self.connection_status = status_msg_text

                if task_success and connected_uri_result and connected_db_name_actual:
                    self.app.current_mongo_uri = connected_uri_result
                    self.app.current_db_name = connected_db_name_actual
                    self.app.available_collections = [
                        item["name"] for item in collections_stats_data
                    ]

                    collections_title.visible = True
                    collections_table.visible = True
                    if collections_stats_data:
                        for coll_data in collections_stats_data:
                            collections_table.add_row(
                                coll_data["name"],
                                str(coll_data.get("count", "N/A")),
                                coll_data.get("avgObjSize", "N/A"),
                                coll_data.get("size", "N/A"),
                                coll_data.get("storageSize", "N/A"),
                                str(coll_data.get("nindexes", "N/A")),
                                key=coll_data["name"],
                            )
                    else:
                        no_coll_msg = "No collections found in this database."
                        if status_msg_text and (
                            "empty" in status_msg_text.plain.lower()
                            or "no collections" in status_msg_text.plain.lower()
                        ):
                            no_coll_msg = status_msg_text.plain
                        collections_table.add_row(no_coll_msg, "", "", "", "", "")

                elif connected_uri_result and connected_db_name_actual:
                    self.app.current_mongo_uri = connected_uri_result
                    self.app.current_db_name = connected_db_name_actual
                    self.app.available_collections = []
                    self.app.active_collection = None
                    collections_title.visible = True
                    collections_table.visible = True
                    collections_table.add_row(
                        status_msg_text.plain if status_msg_text else "Problematic DB context.",
                        "",
                        "",
                        "",
                        "",
                        "",
                    )
                else:
                    self.app.current_mongo_uri = None
                    self.app.current_db_name = None
                    self.app.available_collections = []
                    self.app.active_collection = None
                    collections_title.visible = True
                    collections_table.visible = True
                    collections_table.add_row(
                        status_msg_text.plain if status_msg_text else "Connection failed.",
                        "",
                        "",
                        "",
                        "",
                        "",
                    )

            except WorkerCancelled:
                self.connection_status = Text.from_markup(
                    "[#D08770]Connection task was cancelled.[/]"
                )
            except Exception as e:
                if logger.isEnabledFor(logging.ERROR):
                    logger.error("DB Connection Operation Error in View: %s", e, exc_info=True)
                err_text_display = Text.from_markup(f"[#BF616A]Error: {str(e)[:100]}[/]")
                self.connection_status = err_text_display
                collections_table.clear()
                collections_table.add_row(
                    err_text_display.plain if err_text_display else "Connection Error",
                    "",
                    "",
                    "",
                    "",
                    "",
                )
                collections_title.visible = True
                collections_table.visible = True
                await self.app.push_screen(ErrorDialog("Connection Error", str(e)))

    @on(DataTable.RowSelected, "#collections_data_table")
    async def on_collection_selected_in_table(self, event: DataTable.RowSelected) -> None:
        if event.control.id != "collections_data_table":
            return

        indexes_table = self.query_one("#indexes_data_table", DataTable)
        indexes_title = self.query_one("#indexes_title_label", Label)

        if not event.row_key or not event.row_key.value:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Collection selection event with no row key or value.")
            indexes_table.clear()
            indexes_title.visible = False
            indexes_table.visible = False
            if self.app.active_collection is not None:
                self.app.active_collection = None
            return

        selected_collection_name = str(event.row_key.value)

        if selected_collection_name not in self.app.available_collections:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    "Clicked row '%s' is not an available collection name (e.g., 'Unauthorized' or message row). Ignoring for index loading.",
                    selected_collection_name,
                )

            indexes_table.clear()
            indexes_title.visible = False
            indexes_table.visible = False
            if self.app.active_collection is not None:
                self.app.active_collection = None
            return

        if selected_collection_name == self.app.active_collection:
            return

        self.app.active_collection = selected_collection_name
        await self._load_indexes_for_collection(selected_collection_name)

    async def _load_indexes_for_collection(self, collection_name: str) -> None:
        uri = self.app.current_mongo_uri
        db_name_app = self.app.current_db_name

        indexes_table = self.query_one("#indexes_data_table", DataTable)
        indexes_title = self.query_one("#indexes_title_label", Label)
        indexes_table.clear()

        if not uri or not db_name_app:
            if logger.isEnabledFor(logging.WARNING):
                logger.warning(
                    "Cannot load indexes for '%s': MongoDB not connected or DB name unknown.",
                    collection_name,
                )
            indexes_title.visible = False
            indexes_table.visible = False
            return

        indexes_title.visible = True
        indexes_table.visible = True
        indexes_table.add_row(Text("Loading indexes...", style="italic dim"), "", "", "", "", "")

        try:
            if not core_db_manager.db_connection_active(
                uri=uri, db_name=db_name_app, force_reconnect=False, server_timeout_ms=2000
            ):
                raise ConnectionError(
                    f"Failed to re-verify connection to DB '{db_name_app}' for listing indexes of '{collection_name}'."
                )

            db_instance = core_db_manager.get_mongo_db()

            if db_instance.name != db_name_app:
                if logger.isEnabledFor(logging.ERROR):
                    logger.error(
                        "DB context mismatch. Expected '%s', got '%s' for index listing of '%s'.",
                        db_name_app,
                        db_instance.name,
                        collection_name,
                    )
                raise ConnectionError(f"DB context mismatch. Expected {db_name_app}.")

            collection_obj = db_instance[collection_name]

            def list_indexes_task():
                return list(collection_obj.list_indexes())

            worker: Worker[List[Dict]] = self.app.run_worker(
                list_indexes_task,
                thread=True,
                name=f"list_indexes_{collection_name}",
                group="db_operations",
                description=f"Loading indexes for {collection_name}",
            )
            raw_indexes = await worker.wait()
            indexes_table.clear()

            if worker.is_cancelled:
                indexes_table.add_row(
                    Text("Index loading cancelled by user.", style="italic yellow"),
                    "",
                    "",
                    "",
                    "",
                    "",
                )
                return

            if not raw_indexes:
                indexes_table.add_row(
                    Text(f"No indexes found for '{collection_name}'.", style="italic"),
                    "",
                    "",
                    "",
                    "",
                    "",
                )
                return

            for idx_info in raw_indexes:
                key_dict = idx_info.get("key", {})
                key_str = ", ".join([f"{k}: {v}" for k, v in key_dict.items()])
                other_props_list = []
                excluded_fields = [
                    "v",
                    "key",
                    "name",
                    "ns",
                    "unique",
                    "sparse",
                    "background",
                    "weights",
                    "default_language",
                    "language_override",
                    "textIndexVersion",
                    "2dsphereIndexVersion",
                    "bits",
                    "min",
                    "max",
                    "bucketSize",
                ]
                for p_name, p_val in idx_info.items():
                    if p_name not in excluded_fields:
                        other_props_list.append(
                            f"{p_name}={str(p_val)[:50]}{'...' if len(str(p_val)) > 50 else ''}"
                        )
                other_props_str = ", ".join(other_props_list) if other_props_list else "N/A"
                indexes_table.add_row(
                    idx_info.get("name", "N/A"),
                    key_str,
                    str(idx_info.get("unique", False)),
                    str(idx_info.get("sparse", False)),
                    str(idx_info.get("background", False)),
                    other_props_str,
                )

        except WorkerCancelled:
            indexes_table.clear()
            indexes_table.add_row(
                Text("Index loading was cancelled during operation.", style="italic yellow"),
                "",
                "",
                "",
                "",
                "",
            )
        except PyMongoOperationFailure as e_op:
            logger.warning(
                f"MongoDB operation failure loading indexes for '{collection_name}': {e_op}"
            )
            indexes_table.clear()
            if _is_auth_error(e_op):
                msg = f"Unauthorized to list indexes for '{collection_name}'."
                indexes_table.add_row(Text(msg, style="italic red"), "", "", "", "", "")
                self.app.notify(msg, title="Permission Denied", severity="error", timeout=5)
            else:
                err_details = e_op.details.get("errmsg", str(e_op)) if e_op.details else str(e_op)
                msg = f"Error listing indexes: {str(err_details)[:70]}"
                indexes_table.add_row(Text(msg, style="italic red"), "", "", "", "", "")
                self.app.notify(
                    f"Error listing indexes for '{collection_name}'. Check logs.",
                    title="Operation Error",
                    severity="error",
                    timeout=5,
                )
        except (PyMongoConnectionFailure, ConnectionError) as e_conn:
            logger.error(
                f"Connection error loading indexes for '{collection_name}': {e_conn}", exc_info=True
            )
            indexes_table.clear()
            msg = f"Connection error loading indexes: {str(e_conn)[:70]}"
            indexes_table.add_row(Text(msg, style="italic red"), "", "", "", "", "")
            self.app.notify(
                "Connection error. Please reconnect.",
                title="Connection Error",
                severity="error",
                timeout=5,
            )
        except Exception as e:
            logger.error(
                "Unexpected error loading indexes for collection '%s': %s",
                collection_name,
                e,
                exc_info=True,
            )
            indexes_table.clear()
            indexes_table.add_row(
                Text(f"Unexpected error loading indexes: {str(e)[:70]}", style="italic red"),
                "",
                "",
                "",
                "",
                "",
            )
            await self.app.push_screen(
                ErrorDialog(
                    "Index Load Error", f"Could not load indexes for '{collection_name}': {e!s}"
                )
            )


def _format_bytes_tui(size_bytes: Any) -> str:
    import math

    if size_bytes is None or not isinstance(size_bytes, (int, float)) or size_bytes < 0:
        return "N/A"
    if size_bytes == 0:
        return "0 B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB")
    try:
        if size_bytes <= 0:
            i = 0
        else:
            i = int(math.floor(math.log(size_bytes, 1024)))

        if i >= len(size_name):
            i = len(size_name) - 1
        elif i < 0:
            i = 0
    except ValueError:
        i = 0

    p = math.pow(1024, i)
    s = round(size_bytes / p, 2) if p > 0 else 0
    return f"{s} {size_name[i]}"



import functools
import json
import logging
from typing import Any, Dict, List, Optional, Tuple, Type

import mongo_analyser.core.db as core_db_manager
from mongo_analyser.core import DataExtractor, SchemaAnalyser
from mongo_analyser.dialogs import ErrorDialog
from mongo_analyser.llm_chat import (
    GoogleChat,
    LLMChat,
    OllamaChat,
    OpenAIChat,
)
from mongo_analyser.widgets import ChatMessageList, ChatMessageWidget, LLMConfigPanel
from rich.text import Text
from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.css.query import NoMatches
from textual.widget import Widget
from textual.widgets import (
    Button,
    Input,
    LoadingIndicator,
    Markdown,
    Static,
)
from textual.worker import Worker, WorkerCancelled, WorkerState

logger = logging.getLogger(__name__)


class ChatView(Container):
    current_llm_worker: Worker | None = None
    llm_client_instance: LLMChat | None = None

    ROLE_USER = "user"
    ROLE_AI = "assistant"
    ROLE_SYSTEM = "system"

    PROVIDER_CLASSES: Dict[str, Type[LLMChat]] = {
        "ollama": OllamaChat,
        "openai": OpenAIChat,
        "google": GoogleChat,
    }

    CONTEXT_BLOCK_START_MARKER = "--- START APP-PROVIDED CONTEXT ---\n"
    CONTEXT_BLOCK_END_MARKER = "\n--- END APP-PROVIDED CONTEXT ---\n\n"
    SCHEMA_SECTION_KEY = "schema"
    METADATA_SECTION_KEY = "metadata"
    SAMPLEDOCS_SECTION_KEY = "sample_docs"

    SECTION_TITLE_TEMPLATES = {
        SCHEMA_SECTION_KEY: "[SCHEMA FOR COLLECTION: '{collection_name}']\n",
        METADATA_SECTION_KEY: "[FIELD STATISTICS FOR COLLECTION: '{collection_name}']\n",
        SAMPLEDOCS_SECTION_KEY: "[SAMPLE DOCUMENTS FOR COLLECTION: '{collection_name}'"
        " ({num_docs} docs)]\n",
    }
    DEFAULT_SAMPLE_DOCS_COUNT = 3

    _current_ai_message_widget: Optional[ChatMessageWidget] = None
    _full_response_content: str = ""

    def __init__(
        self,
        *children: Widget,
        name: Optional[str] = None,
        id: Optional[str] = None,
        classes: Optional[str] = None,
        disabled: bool = False,
    ):
        super().__init__(
            *children,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.chat_history: List[Dict[str, str]] = []
        self.active_contexts: Dict[str, Dict[str, Any]] = {}

    def _log_chat_message(self, role: str, message_content: str) -> None:
        try:
            chat_list_widget = self.query_one("#chat_log_widget", ChatMessageList)
            chat_list_widget.add_message(role, message_content)
            chat_list_widget.scroll_end(animate=False)
        except NoMatches:
            logger.warning("Chat log widget (#chat_log_widget) not found for logging message.")
        except Exception as e:
            logger.error(f"Error logging chat message: {e}", exc_info=True)

    def _update_chat_status_line(
        self, status: str = "Idle", current_messages: int | None = None
    ) -> None:
        try:
            panel = self.query_one(LLMConfigPanel)
            chat_status_widget = self.query_one("#chat_status_line", Static)
            prov = panel.provider or ""
            mdl = panel.model or ""
            provider_display = prov.capitalize() if prov else "N/A"
            model_display = mdl if mdl else "N/A"
            max_hist = panel.max_history_messages
            if max_hist == -1:
                max_hist_display = "None"
            elif max_hist == 0:
                max_hist_display = "All"
            else:
                max_hist_display = str(max_hist)
            actual_hist_len = len(self._get_effective_history_for_llm())
            history_info = f"History: {actual_hist_len} (max: {max_hist_display})"
            chat_status_widget.update(
                f"Provider: {provider_display} | Model: {model_display} | {history_info} | Status: {status}"
            )
        except NoMatches:
            logger.warning("ChatView: Could not update chat status line (widget not found).")
        except Exception as e:
            logger.error(f"Error updating chat status line: {e}", exc_info=True)

    def _reset_chat_log_and_status(self, status_message: str = "New session started.") -> None:
        self.chat_history.clear()
        self.active_contexts.clear()
        try:
            log_widget = self.query_one("#chat_log_widget", ChatMessageList)
            log_widget.clear_messages()
        except NoMatches:
            logger.warning("Chat log widget not found for clearing during reset.")

        self._apply_contexts_to_input_field()
        self._log_chat_message(self.ROLE_SYSTEM, status_message)
        self._update_chat_status_line(status="Idle", current_messages=0)

    def on_mount(self) -> None:
        self.chat_history: List[Dict[str, str]] = []
        self.active_contexts: Dict[str, Dict[str, Any]] = {}
        self.llm_client_instance = None
        self._reset_chat_log_and_status("ChatView initialized. Configure LLM in sidebar.")
        self.focus_default_widget()
        try:
            self.query_one("#chat_model_loading_indicator", LoadingIndicator).display = False
            self.query_one("#active_context_indicator_label", Static).display = False
        except NoMatches:
            logger.warning(
                "ChatView: Initial display setup for indicators failed (widget not found)."
            )

    def compose(self) -> ComposeResult:
        with Horizontal(id="chat_interface_horizontal_layout"):
            with Vertical(id="chat_main_area", classes="chat_column_main"):
                yield Static(
                    "Provider: N/A | Model: N/A | History: 0 (max: N/A) | Status: Idle",
                    id="chat_status_line",
                    classes="chat_status",
                )
                yield LoadingIndicator(id="chat_model_loading_indicator")
                yield ChatMessageList(id="chat_log_widget")
                with Horizontal(classes="chat_action_buttons"):
                    yield Button(
                        "Prepend Collection Schema",
                        id="prepend_schema_button",
                        classes="context_button",
                    )
                    yield Button(
                        "Prepend Collection Metadata",
                        id="prepend_metadata_button",
                        classes="context_button",
                    )
                    yield Button(
                        "Prepend Sample Docs",
                        id="prepend_sample_docs_button",
                        classes="context_button",
                    )
                yield Static("", id="active_context_indicator_label", classes="context_indicator")
                with Horizontal(id="chat_input_bar", classes="chat_input_container"):
                    yield Input(placeholder="Type a message...", id="chat_message_input")
                    yield Button(
                        "Send",
                        id="send_chat_message_button",
                        variant="primary",
                        classes="chat_button",
                    )
                    yield Button(
                        "Stop", id="stop_chat_message_button", classes="chat_button", disabled=True
                    )
            yield LLMConfigPanel(id="chat_llm_config_panel", classes="chat_column_sidebar")

    def focus_default_widget(self) -> None:
        try:
            self.query_one("#chat_message_input", Input).focus()
        except NoMatches:
            logger.debug("ChatView: Could not focus default input (#chat_message_input).")

    @on(LLMConfigPanel.ProviderChanged)
    async def handle_provider_change_from_llm_config_panel(
        self, event: LLMConfigPanel.ProviderChanged
    ) -> None:
        logger.info(f"ChatView: ProviderChanged event with provider: {event.provider}")
        self.llm_client_instance = None
        self._log_chat_message(self.ROLE_SYSTEM, "Provider changed. LLM client reset.")
        if event.provider:
            try:
                panel = self.query_one(LLMConfigPanel)
                if panel.provider != event.provider:
                    panel.provider = event.provider
            except NoMatches:
                pass
            self.app.call_later(self._load_models_for_provider, event.provider)
        else:
            try:
                panel = self.query_one(LLMConfigPanel)
                panel.update_models_list([], "Select Provider First")
                panel.model = None
            except NoMatches:
                logger.error(
                    "ChatView: LLMConfigPanel not found when handling null provider change."
                )
            self._update_chat_status_line(status="Provider cleared")

    async def _load_models_for_provider(self, provider_value: str) -> None:
        loader = self.query_one("#chat_model_loading_indicator", LoadingIndicator)
        try:
            panel = self.query_one(LLMConfigPanel)
        except NoMatches:
            logger.error("ChatView: LLMConfigPanel not found when loading models.")
            self._log_chat_message(self.ROLE_SYSTEM, "Internal error: LLM config panel missing.")
            self._update_chat_status_line(status="Panel Error")
            if loader.is_mounted:
                loader.display = False
            return

        llm_class = self.PROVIDER_CLASSES.get(provider_value)
        if not llm_class:
            panel.update_models_list([], f"Unknown provider: {provider_value}")
            if panel.model is not None:
                panel.model = None
            self._log_chat_message(
                self.ROLE_SYSTEM, f"Cannot load models for unknown provider: {provider_value}"
            )
            if loader.is_mounted:
                loader.display = False
            return

        panel.set_model_select_loading(True, f"Loading models for {provider_value}...")
        self._update_chat_status_line(status=f"Loading {provider_value} models...")
        self._log_chat_message(self.ROLE_SYSTEM, f"Fetching models for {provider_value}...")
        if loader.is_mounted:
            loader.display = True

        listed: List[str] = []
        error: Optional[str] = None

        client_cfg_from_panel = panel.get_llm_config()
        client_cfg_from_panel["provider_hint"] = provider_value

        try:
            worker: Worker[List[str]] = self.app.run_worker(
                functools.partial(llm_class.list_models, client_config=client_cfg_from_panel),
                thread=True,
                group="model_listing",
            )
            listed = await worker.wait()
            if worker.is_cancelled:
                error = "Model loading cancelled by worker."
        except WorkerCancelled:
            error = "Model loading cancelled."
        except Exception as e:
            logger.error(f"ChatView: Error listing models for {provider_value}: {e}", exc_info=True)
            error = f"Failed to list models: {e.__class__.__name__}: {str(e)[:60]}"

        if loader.is_mounted:
            loader.display = False
        panel.set_model_select_loading(False)

        current_status_after_listing = "Models loaded"
        if error:
            panel.update_models_list([], error)
            self._log_chat_message(self.ROLE_SYSTEM, error)
            current_status_after_listing = "Model list error"
            if panel.model is not None:
                panel.model = None
        else:
            options = [(m, m) for m in listed]
            prompt_if_empty = "No models found for this provider." if not listed else "Select Model"
            panel.update_models_list(options, prompt_if_empty)
            current_status_after_listing = "Models loaded" if listed else "No models found"

            selected_model_for_panel: Optional[str] = None
            if listed:
                default_configured_model_key = f"llm_default_model_{provider_value}"
                default_configured_model = self.app.config_manager.get_setting(
                    default_configured_model_key
                )

                if default_configured_model and default_configured_model in listed:
                    selected_model_for_panel = default_configured_model
                    logger.info(
                        f"Using configured default model for {provider_value}: {selected_model_for_panel}"
                    )
                else:
                    if default_configured_model:
                        logger.warning(
                            f"Configured default model '{default_configured_model}' for {provider_value} not found in listed models. Using first available."
                        )
                    selected_model_for_panel = listed[0]
                    logger.info(
                        f"Using first available model for {provider_value}: {selected_model_for_panel}"
                    )

            if panel.model != selected_model_for_panel:
                panel.model = selected_model_for_panel
            elif selected_model_for_panel is None and panel.model is not None:
                panel.model = None
            elif selected_model_for_panel is not None and panel.model is None:
                panel.model = selected_model_for_panel

        self._update_chat_status_line(status=current_status_after_listing)
        if panel.model is None and (error or not listed):
            await self.handle_model_change_from_llm_config_panel(LLMConfigPanel.ModelChanged(None))
        elif panel.model is not None:
            if (
                self.llm_client_instance is None
                or self.llm_client_instance.model_name != panel.model
            ):
                await self.handle_model_change_from_llm_config_panel(
                    LLMConfigPanel.ModelChanged(panel.model)
                )

    @on(LLMConfigPanel.ModelChanged)
    async def handle_model_change_from_llm_config_panel(
        self, event: LLMConfigPanel.ModelChanged
    ) -> None:
        model_value = event.model
        logger.info(
            f"ChatView: handle_model_change_from_llm_config_panel received ModelChanged"
            f" event with model: {model_value}"
        )
        self.active_contexts.clear()
        if model_value:
            self._reset_chat_log_and_status(f"Model set to: {model_value}. Session reset.")
            try:
                self.query_one("#chat_message_input", Input).value = ""
            except NoMatches:
                pass
            self._apply_contexts_to_input_field()
            if self._create_and_set_llm_client():
                self._log_chat_message(self.ROLE_SYSTEM, "Session ready. LLM client configured.")
                self._update_chat_status_line(status="Ready")
            else:
                self._log_chat_message(
                    self.ROLE_SYSTEM,
                    "Client configuration error for selected model. Check logs for details.",
                )
                self._update_chat_status_line(status="Client Error")
            self.focus_default_widget()
        else:
            self.llm_client_instance = None
            self._reset_chat_log_and_status("No model selected or available.")
            self._log_chat_message(
                self.ROLE_SYSTEM, "Model deselected or unavailable. LLM client cleared."
            )
            self._update_chat_status_line(status="Select model")
            self._apply_contexts_to_input_field()

    def _create_and_set_llm_client(self) -> bool:
        try:
            panel = self.query_one(LLMConfigPanel)
        except NoMatches:
            logger.error("ChatView: LLMConfigPanel not found during client creation.")
            self._log_chat_message(self.ROLE_SYSTEM, "LLM Configuration panel not found.")
            return False
        cfg = panel.get_llm_config()
        provider = cfg.get("provider_hint")
        model_name = cfg.get("model_name")

        if not provider or not model_name:
            self._log_chat_message(
                self.ROLE_SYSTEM, "Provider or model not properly selected for client creation."
            )
            return False
        llm_class = self.PROVIDER_CLASSES.get(provider)
        if not llm_class:
            self._log_chat_message(
                self.ROLE_SYSTEM, f"Unknown provider '{provider}' for client creation."
            )
            return False

        client_kwargs = {"model_name": model_name}
        temperature = cfg.get(
            "temperature", self.app.config_manager.get_setting("llm_default_temperature")
        )

        if provider == "ollama":
            client_kwargs.update(
                {
                    k: v
                    for k, v in cfg.items()
                    if k
                    not in ["provider_hint", "model_name", "temperature", "max_history_messages"]
                }
            )
            client_kwargs.setdefault("options", {})
            if temperature is not None:
                client_kwargs["options"]["temperature"] = temperature

        elif provider == "openai":
            client_kwargs.update(
                {
                    k: v
                    for k, v in cfg.items()
                    if k not in ["provider_hint", "model_name", "max_history_messages"]
                }
            )
            if temperature is not None:
                client_kwargs["temperature"] = temperature

        elif provider == "google":
            client_kwargs.update(
                {
                    k: v
                    for k, v in cfg.items()
                    if k
                    not in ["provider_hint", "model_name", "temperature", "max_history_messages"]
                }
            )
            if temperature is not None:
                client_kwargs["generation_config"] = {"temperature": temperature}

        logger.debug(
            f"ChatView: Instantiating LLM class '{llm_class.__name__}' "
            f"with effective kwargs: {client_kwargs}"
        )
        try:
            new_client = llm_class(**client_kwargs)
            self.llm_client_instance = new_client
            logger.info(f"ChatView: LLM client successfully created for {provider}:{model_name}.")
            return True
        except Exception as e:
            logger.error(
                f"ChatView: Failed to create LLM client for {provider}:{model_name}."
                f" KWARGS: {client_kwargs}. Error: {e}",
                exc_info=True,
            )
            self._log_chat_message(
                self.ROLE_SYSTEM,
                f"Error creating LLM client for '{model_name}':"
                f" {e.__class__.__name__} - {str(e)[:100]}. See console log for full traceback.",
            )
            self.llm_client_instance = None
            return False

    def _get_effective_history_for_llm(self) -> List[Dict[str, str]]:
        hist = [m for m in self.chat_history if m["role"] in {self.ROLE_USER, self.ROLE_AI}]
        try:
            panel = self.query_one(LLMConfigPanel)
            max_hist = panel.max_history_messages
            if max_hist == -1:
                return []
            if max_hist is not None and max_hist == 0:
                return hist
            if max_hist is not None and 0 < max_hist < len(hist):
                return hist[-max_hist:]
        except NoMatches:
            logger.warning("LLMConfigPanel not found in _get_effective_history_for_llm")
        return hist

    def _prepare_for_stream(self) -> None:
        self._full_response_content = ""
        try:
            chat_list_widget = self.query_one("#chat_log_widget", ChatMessageList)
            self._current_ai_message_widget = ChatMessageWidget(self.ROLE_AI, "●")
            chat_list_widget.mount(self._current_ai_message_widget)
            chat_list_widget.scroll_end(animate=False)
        except NoMatches:
            logger.error("ChatView: Could not find #chat_log_widget to prepare for stream.")
            self._current_ai_message_widget = None

    def _handle_stream_chunk(self, chunk: str) -> None:
        if self._current_ai_message_widget and self._current_ai_message_widget.is_mounted:
            self._full_response_content += chunk
            cursor_char = " ●" if len(self._full_response_content) % 10 < 5 else "●"
            markdown_widget = self._current_ai_message_widget.query_one(Markdown)
            markdown_widget.update(self._full_response_content + cursor_char)
            try:
                self.query_one("#chat_log_widget", ChatMessageList).scroll_end(animate=False)
            except NoMatches:
                pass

    def _handle_stream_end(self, cancelled: bool = False) -> None:
        final_text = self._full_response_content.strip()

        if self.current_llm_worker and self.current_llm_worker.state == WorkerState.CANCELLED:
            cancelled = True

        if cancelled and not final_text.endswith("[Stopped by user]"):
            final_text += "\n[Stopped by user]"

        if self._current_ai_message_widget and self._current_ai_message_widget.is_mounted:
            self._current_ai_message_widget.query_one(Markdown).update(final_text)

        self.chat_history.append({"role": self.ROLE_AI, "content": final_text})
        self._cleanup_after_send()

    def _handle_stream_error(self, error: Exception) -> None:
        logger.error(f"Error in streaming worker task: {error}", exc_info=True)
        err_msg = f"LLM Error: {error!s}"
        if self._current_ai_message_widget and self._current_ai_message_widget.is_mounted:
            self._current_ai_message_widget.query_one(Markdown).update(err_msg)
        else:
            self._log_chat_message(self.ROLE_SYSTEM, err_msg)

        self.chat_history.append({"role": self.ROLE_SYSTEM, "content": err_msg})
        self._cleanup_after_send()

    def _cleanup_after_send(self) -> None:
        if self.is_mounted:
            try:
                input_widget = self.query_one("#chat_message_input", Input)
                send_btn = self.query_one("#send_chat_message_button", Button)
                stop_btn = self.query_one("#stop_chat_message_button", Button)

                input_widget.disabled = False
                send_btn.disabled = False
                stop_btn.disabled = True
                self._update_chat_status_line(status="Ready")
                input_widget.focus()
            except NoMatches:
                logger.warning("ChatView: UI elements not found in _cleanup_after_send.")

        self._current_ai_message_widget = None
        self.current_llm_worker = None

    def _streaming_worker_task(
        self, client: LLMChat, message: str, history: List[Dict[str, str]]
    ) -> None:
        """The actual function run by the worker to stream messages."""
        try:
            for chunk in client.stream_message(message=message, history=history):
                if self.current_llm_worker and self.current_llm_worker.is_cancelled:
                    self.app.call_from_thread(self._handle_stream_end, cancelled=True)
                    return
                self.app.call_from_thread(self._handle_stream_chunk, chunk)

            self.app.call_from_thread(self._handle_stream_end, cancelled=False)
        except Exception as e:
            logger.error(f"Exception in _streaming_worker_task: {e}", exc_info=True)
            self.app.call_from_thread(self._handle_stream_error, e)

    async def _send_user_message(self) -> None:
        try:
            input_widget = self.query_one("#chat_message_input", Input)
            send_btn = self.query_one("#send_chat_message_button", Button)
            stop_btn = self.query_one("#stop_chat_message_button", Button)
        except NoMatches:
            await self.app.push_screen(ErrorDialog("UI Error", "Chat UI elements missing."))
            return

        if not self.llm_client_instance:
            await self.app.push_screen(
                ErrorDialog("LLM Error", "LLM not configured. Please select a provider and model.")
            )
            return

        user_typed_message = input_widget.value.strip()
        if not user_typed_message:
            self.app.notify("Cannot send an empty message.", title="Input Error", severity="error")
            return

        self._log_chat_message(self.ROLE_USER, user_typed_message)
        self.chat_history.append({"role": self.ROLE_USER, "content": user_typed_message})

        history_for_llm = self._get_effective_history_for_llm()
        context_block_content = self._build_context_block_string()
        message_for_llm = context_block_content + user_typed_message

        input_widget.value = ""
        self.active_contexts.clear()
        self._apply_contexts_to_input_field()

        input_widget.disabled = True
        send_btn.disabled = True
        stop_btn.disabled = False
        self._update_chat_status_line(status="Receiving...")

        self._prepare_for_stream()

        client = self.llm_client_instance

        task = functools.partial(
            self._streaming_worker_task,
            client=client,
            message=message_for_llm,
            history=history_for_llm,
        )

        if self.current_llm_worker and self.current_llm_worker.state == WorkerState.RUNNING:
            self.current_llm_worker.cancel()

        self.current_llm_worker = self.app.run_worker(
            task, thread=True, group="llm_call_stream", exclusive=True
        )

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        bid = event.button.id
        if bid == "send_chat_message_button":
            await self._send_user_message()
        elif bid == "stop_chat_message_button":
            if self.current_llm_worker and self.current_llm_worker.state == WorkerState.RUNNING:
                logger.info("Stop button pressed. Cancelling LLM worker.")
                self.current_llm_worker.cancel()

                self._log_chat_message(self.ROLE_SYSTEM, "Attempting to stop LLM response...")
                try:
                    self.query_one("#stop_chat_message_button", Button).disabled = True
                except NoMatches:
                    pass

        elif bid == "prepend_schema_button":
            await self._handle_prepend_context(self.SCHEMA_SECTION_KEY, event.button)
        elif bid == "prepend_metadata_button":
            await self._handle_prepend_context(self.METADATA_SECTION_KEY, event.button)
        elif bid == "prepend_sample_docs_button":
            await self._handle_prepend_context(self.SAMPLEDOCS_SECTION_KEY, event.button)

    async def _handle_prepend_context(self, context_type_key: str, button: Button) -> None:
        coll = self.app.active_collection
        if not coll:
            self.app.notify(
                "No active collection selected to prepend context.",
                title="Context Error",
                severity="warning",
            )
            return

        content_str: Optional[str] = None
        num_docs_for_title: Optional[int] = None
        original_button_label = button.label
        button.label = Text("Loading...", style="italic yellow")
        button.disabled = True

        try:
            if context_type_key == self.SCHEMA_SECTION_KEY:
                content_str = await self._get_active_collection_hierarchical_schema()
            elif context_type_key == self.METADATA_SECTION_KEY:
                content_str = await self._get_active_collection_field_stats()
            elif context_type_key == self.SAMPLEDOCS_SECTION_KEY:
                num_docs_for_title = self.DEFAULT_SAMPLE_DOCS_COUNT
                content_str = await self._get_active_collection_sample_docs(
                    num_docs=num_docs_for_title
                )

            if content_str:
                self._set_context_for_input(
                    context_type_key, coll, content_str, num_docs=num_docs_for_title
                )
                self._log_chat_message(
                    self.ROLE_SYSTEM,
                    f"{context_type_key.replace('_', ' ').capitalize()}"
                    f" for collection '{coll}' prepared for input.",
                )
                self.app.notify(
                    f"{context_type_key.replace('_', ' ').capitalize()}"
                    f" for '{coll}' added/updated for next message.",
                    title="Context Added",
                )
                button.label = Text(
                    f"{context_type_key.replace('_', ' ').capitalize()} Added ✓",
                    style="italic green",
                )
            else:
                self.app.notify(
                    f"Could not retrieve {context_type_key.replace('_', ' ')} for '{coll}'.",
                    title="Context Error",
                    severity="error",
                )
                button.label = original_button_label

        except Exception as e:
            logger.error(
                f"Error handling prepend context for {context_type_key}: {e}", exc_info=True
            )
            self.app.notify(f"Error: {e}", title="Context Preparation Error", severity="error")
            button.label = original_button_label
        finally:

            def revert_button():
                if button.is_mounted and button.label != original_button_label:
                    if "Added ✓" in button.label.plain:
                        self.set_timer(1.5, lambda: setattr(button, "label", original_button_label))
                    else:
                        button.label = original_button_label
                button.disabled = False

            if "Loading..." in button.label.plain:
                button.label = original_button_label
                button.disabled = False
            else:
                self.set_timer(2.5, revert_button)

    async def _fetch_schema_and_stats_if_needed(
        self, collection_name: str
    ) -> Tuple[Optional[Dict], Optional[Dict]]:
        if not (self.app.current_mongo_uri and self.app.current_db_name):
            return None, None

        cached_results = self.app.current_schema_analysis_results
        if (
            isinstance(cached_results, dict)
            and cached_results.get("collection_name") == collection_name
            and "hierarchical_schema" in cached_results
            and "field_stats" in cached_results
        ):
            logger.info(f"Using cached schema and stats for '{collection_name}'.")
            return cached_results["hierarchical_schema"], cached_results["field_stats"]

        logger.info(f"No cached results for '{collection_name}' or cache mismatch. Fetching live.")
        self.app.notify(f"Fetching schema/stats for '{collection_name}'...", title="Data Fetch")

        try:
            if not core_db_manager.db_connection_active(
                uri=self.app.current_mongo_uri,
                db_name=self.app.current_db_name,
                server_timeout_ms=3000,
            ):
                raise ConnectionError("DB connection lost or could not be re-established.")

            pymongo_collection = SchemaAnalyser.get_collection(
                self.app.current_mongo_uri, self.app.current_db_name, collection_name
            )

            analysis_task = functools.partial(
                SchemaAnalyser.infer_schema_and_field_stats,
                collection=pymongo_collection,
                sample_size=100,
            )
            worker: Worker[Tuple[Dict, Dict]] = self.app.run_worker(
                analysis_task, thread=True, group="chat_context_fetch"
            )
            flat_schema_data, field_stats_data = await worker.wait()

            if worker.is_cancelled:
                self.app.notify(
                    f"Schema/stats fetch for '{collection_name}' cancelled.",
                    title="Fetch Cancelled",
                )
                return None, None

            if not flat_schema_data and not field_stats_data:
                self.app.notify(
                    f"No schema/stats data returned for '{collection_name}'.", title="Fetch Error"
                )
                return None, None

            hierarchical_schema = SchemaAnalyser.schema_to_hierarchical(flat_schema_data or {})

            self.app.current_schema_analysis_results = {
                "flat_schema": flat_schema_data,
                "field_stats": field_stats_data,
                "hierarchical_schema": hierarchical_schema,
                "collection_name": collection_name,
            }
            return hierarchical_schema, field_stats_data

        except Exception as e:
            logger.error(f"Error fetching schema/stats for injection: {e}", exc_info=True)
            await self.app.push_screen(
                ErrorDialog(
                    "Data Fetch Error", f"Could not fetch data for '{collection_name}': {e!s}"
                )
            )
            return None, None

    async def _get_active_collection_hierarchical_schema(self) -> Optional[str]:
        coll = self.app.active_collection
        if not coll:
            return None
        hierarchical_schema, _ = await self._fetch_schema_and_stats_if_needed(coll)
        if hierarchical_schema is None:
            return None
        try:
            return json.dumps(hierarchical_schema, indent=2, default=str)
        except TypeError:
            logger.error(f"Hierarchical schema for '{coll}' is not JSON serializable.")
            return str(hierarchical_schema)

    async def _get_active_collection_field_stats(self) -> Optional[str]:
        coll = self.app.active_collection
        if not coll:
            return None
        _, field_stats = await self._fetch_schema_and_stats_if_needed(coll)
        if field_stats is None:
            return None
        try:
            return json.dumps(field_stats, indent=2, default=str)
        except TypeError:
            logger.error(f"Field stats for '{coll}' are not JSON serializable.")
            return str(field_stats)

    async def _get_active_collection_sample_docs(self, num_docs: int) -> Optional[str]:
        coll = self.app.active_collection
        if not (self.app.current_mongo_uri and self.app.current_db_name and coll):
            return None

        self.app.notify(f"Fetching {num_docs} sample docs for '{coll}'...", title="Data Fetch")
        try:
            if not core_db_manager.db_connection_active(
                uri=self.app.current_mongo_uri,
                db_name=self.app.current_db_name,
                server_timeout_ms=3000,
            ):
                raise ConnectionError("DB connection lost for sample doc fetching.")

            fetch_task = functools.partial(
                DataExtractor.get_newest_documents,
                self.app.current_mongo_uri,
                self.app.current_db_name,
                coll,
                num_docs,
            )
            worker: Worker[List[Dict]] = self.app.run_worker(
                fetch_task, thread=True, group="chat_sample_docs_fetch"
            )
            documents = await worker.wait()

            if worker.is_cancelled:
                self.app.notify(
                    f"Sample doc fetch for '{coll}' cancelled.", title="Fetch Cancelled"
                )
                return None

            if not documents:
                self.app.notify(f"No sample documents found for '{coll}'.", title="Data Fetch")
                return "[]"

            return json.dumps(documents, indent=2, default=str)

        except Exception as e:
            logger.error(f"Error fetching sample documents for injection: {e}", exc_info=True)
            await self.app.push_screen(
                ErrorDialog(
                    "Sample Docs Fetch Error", f"Could not fetch samples for '{coll}': {e!s}"
                )
            )
            return None

    def _set_context_for_input(
        self,
        context_type_key: str,
        collection_name: str,
        content: str,
        num_docs: Optional[int] = None,
    ) -> None:
        self.active_contexts[context_type_key] = {
            "collection": collection_name,
            "content": content,
            "num_docs": num_docs,
        }
        self._apply_contexts_to_input_field()

    def _build_context_block_string(self) -> str:
        if not self.active_contexts:
            return ""

        block_content_parts = []
        ordered_keys = [
            self.SCHEMA_SECTION_KEY,
            self.METADATA_SECTION_KEY,
            self.SAMPLEDOCS_SECTION_KEY,
        ]

        for key in ordered_keys:
            if key in self.active_contexts:
                context_item = self.active_contexts[key]
                title_template = self.SECTION_TITLE_TEMPLATES[key]
                title_args = {"collection_name": context_item["collection"]}
                if key == self.SAMPLEDOCS_SECTION_KEY and context_item.get("num_docs") is not None:
                    title_args["num_docs"] = context_item["num_docs"]
                section_title = title_template.format(**title_args)
                block_content_parts.append(
                    f"{section_title}```json\n{context_item['content']}\n```\n"
                )
        if not block_content_parts:
            return ""
        return (
            self.CONTEXT_BLOCK_START_MARKER
            + "\n".join(block_content_parts)
            + self.CONTEXT_BLOCK_END_MARKER
        )

    def _update_active_context_indicator(self) -> None:
        try:
            indicator_label = self.query_one("#active_context_indicator_label", Static)
            if not self.active_contexts:
                indicator_label.update("")
                indicator_label.display = False
                return

            active_items = []
            ordered_keys = [
                self.SCHEMA_SECTION_KEY,
                self.METADATA_SECTION_KEY,
                self.SAMPLEDOCS_SECTION_KEY,
            ]
            for key in ordered_keys:
                if key in self.active_contexts:
                    item = self.active_contexts[key]
                    coll_name = item["collection"]
                    key_display = key.replace("_", " ").capitalize()
                    if key == self.SAMPLEDOCS_SECTION_KEY and item.get("num_docs") is not None:
                        active_items.append(f"{key_display} ({item['num_docs']}) for '{coll_name}'")
                    else:
                        active_items.append(f"{key_display} for '{coll_name}'")

            if active_items:
                indicator_text = "Prepending: " + "; ".join(active_items)
                indicator_label.update(Text(indicator_text, style="italic dim"))
                indicator_label.display = True
            else:
                indicator_label.update("")
                indicator_label.display = False
        except NoMatches:
            logger.warning("Active context indicator label not found.")

    def _apply_contexts_to_input_field(self) -> None:
        self._update_active_context_indicator()
        try:
            input_widget = self.query_one("#chat_message_input", Input)
            if self.app.focused is input_widget:
                input_widget.action_end()
        except NoMatches:
            logger.warning("Chat input widget not found in _apply_contexts_to_input_field.")

    @on(LLMConfigPanel.NewSessionRequested)
    async def handle_new_session_requested_from_llm_config_panel(
        self,
    ) -> None:
        logger.info("ChatView: NewSessionRequested event received.")
        self.active_contexts.clear()
        try:
            panel = self.query_one(LLMConfigPanel)
        except NoMatches:
            await self.app.push_screen(ErrorDialog("UI Error", "LLM Config panel not found."))
            return
        current_model_on_panel = panel.model
        if not panel.provider or not current_model_on_panel:
            self._reset_chat_log_and_status("Select provider and model first.")
            await self.app.push_screen(
                ErrorDialog(
                    "Configuration Incomplete",
                    "Please select a provider and model before starting a new session.",
                )
            )
            return
        self._reset_chat_log_and_status("New session started. LLM (re)configuring...")
        try:
            self.query_one("#chat_message_input", Input).value = ""
        except NoMatches:
            pass
        self._apply_contexts_to_input_field()
        if self._create_and_set_llm_client():
            self._log_chat_message(self.ROLE_SYSTEM, "LLM client (re)configured for new session.")
            self._update_chat_status_line(status="Ready")
        else:
            self._log_chat_message(
                self.ROLE_SYSTEM, "Client Error during new session setup. Check logs."
            )
            self._update_chat_status_line(status="Client Error")
        self.focus_default_widget()

    async def on_input_submitted(self, event: Input.Submitted) -> None:
        if event.input.id == "chat_message_input":
            try:
                send_button = self.query_one("#send_chat_message_button", Button)
                if not send_button.disabled:
                    await self._send_user_message()
            except NoMatches:
                logger.error("Send button not found on input submission.")



from textual.containers import VerticalScroll

from .chat_message_widget import ChatMessageWidget


class ChatMessageList(VerticalScroll):
    def add_message(self, role: str, content: str) -> None:
        message_widget = ChatMessageWidget(role, content)
        self.mount(message_widget)
        self.scroll_end(animate=True)

    def clear_messages(self) -> None:
        self.query(ChatMessageWidget).remove()



from textual.app import ComposeResult
from textual.containers import Vertical
from textual.widgets import DataTable, Label


class CollectionStatsWidget(Vertical):
    def compose(self) -> ComposeResult:
        yield Label("Collection Statistics")
        yield DataTable(id="internal_collection_stats_table")

    def on_mount(self):
        table = self.query_one(DataTable)
        table.add_columns("Name", "Docs", "Avg Size", "Total Size", "Storage Size")



import logging
from typing import Any, Dict, List, Optional, Tuple

from textual.app import ComposeResult
from textual.containers import VerticalScroll
from textual.css.query import NoMatches
from textual.message import Message
from textual.reactive import reactive
from textual.widgets import Button, Input, Label, Select

logger = logging.getLogger(__name__)


class LLMConfigPanel(VerticalScroll):
    DEFAULT_TEMPERATURE = 0.7
    DEFAULT_MAX_HISTORY = 20

    class ProviderChanged(Message):
        def __init__(self, provider: Optional[str]):
            self.provider = provider
            super().__init__()

    class ModelChanged(Message):
        def __init__(self, model: Optional[str]):
            self.model = model
            super().__init__()

    class NewSessionRequested(Message):
        pass

    provider: reactive[Optional[str]] = reactive(None)
    model: reactive[Optional[str]] = reactive(None)
    temperature: reactive[Optional[float]] = reactive(DEFAULT_TEMPERATURE)
    max_history_messages: reactive[Optional[int]] = reactive(DEFAULT_MAX_HISTORY)

    def compose(self) -> ComposeResult:
        yield Label("Session Config", classes="panel_title")
        yield Label("Provider:")
        yield Select(
            [("Ollama", "ollama"), ("OpenAI", "openai"), ("Google", "google")],
            prompt="Select Provider",
            id="llm_config_provider_select",
            allow_blank=False,
            value="ollama",
        )
        yield Label("Model:")
        yield Select(
            [],
            prompt="Select Provider First",
            id="llm_config_model_select",
            allow_blank=True,
            value=Select.BLANK,
        )

        yield Label(f"Temperature (default: {self.DEFAULT_TEMPERATURE}):")
        yield Input(
            placeholder=str(self.DEFAULT_TEMPERATURE),
            id="llm_config_temperature",
            value=str(self.DEFAULT_TEMPERATURE),
            tooltip="Controls randomness (0.0-1.0). Higher values (e.g., 0.9) for more creative and"
            " diverse responses, lower (e.g., 0.2) for more deterministic/focused ones."
            f" Default: {self.DEFAULT_TEMPERATURE}",
        )

        yield Label(f"Max History (0=all, -1=none, default: {self.DEFAULT_MAX_HISTORY}):")
        yield Input(
            placeholder=str(self.DEFAULT_MAX_HISTORY),
            id="llm_config_max_history",
            value=str(self.DEFAULT_MAX_HISTORY),
            tooltip="Number of recent conversation turns (user+AI message pairs) to include as context."
            " '0' includes all available history. '-1' includes no history (current message only)."
            f" Default: {self.DEFAULT_MAX_HISTORY}",
        )

        yield Button("New Chat Session", id="llm_config_new_session_button", variant="primary")

    def on_mount(self) -> None:
        self.provider = None
        select = self.query_one("#llm_config_provider_select", Select)
        select.value = "ollama"
        ms = self.query_one("#llm_config_model_select", Select)
        ms.disabled = True
        ms.value = Select.BLANK
        self._update_temperature()
        self._update_max_history()
        logger.info("LLMConfigPanel: on_mount complete; loading default provider models.")

    async def on_select_changed(self, event: Select.Changed) -> None:
        if event.select.id == "llm_config_provider_select":
            new_provider_value = str(event.value)
            if self.provider != new_provider_value:
                self.provider = new_provider_value
                try:
                    model_select = self.query_one("#llm_config_model_select", Select)
                    model_select.set_options([])
                    model_select.disabled = True
                    model_select.prompt = "Loading models…"
                    model_select.value = Select.BLANK
                    self.model = None
                except NoMatches:
                    logger.warning("LLMConfigPanel: Model select not found during provider change.")
                self.post_message(self.ProviderChanged(self.provider))
        elif event.select.id == "llm_config_model_select":
            new_model_value = str(event.value) if event.value != Select.BLANK else None
            if self.model != new_model_value:
                self.model = new_model_value
                self.post_message(self.ModelChanged(self.model))

    async def on_input_changed(self, event: Input.Changed) -> None:
        if event.input.id == "llm_config_temperature":
            self._update_temperature()
        elif event.input.id == "llm_config_max_history":
            self._update_max_history()

    def _update_temperature(self) -> None:
        try:
            temp_input = self.query_one("#llm_config_temperature", Input)
            self.temperature = float(temp_input.value)
        except ValueError:
            self.temperature = self.DEFAULT_TEMPERATURE
        except NoMatches:
            logger.error("LLMConfigPanel: Temperature input not found.")
            self.temperature = self.DEFAULT_TEMPERATURE

    def _update_max_history(self) -> None:
        try:
            history_input = self.query_one("#llm_config_max_history", Input)
            self.max_history_messages = int(history_input.value)
        except ValueError:
            self.max_history_messages = self.DEFAULT_MAX_HISTORY
        except NoMatches:
            logger.error("LLMConfigPanel: Max history input not found.")
            self.max_history_messages = self.DEFAULT_MAX_HISTORY

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "llm_config_new_session_button":
            self.post_message(self.NewSessionRequested())

    def update_models_list(self, models: List[Tuple[str, str]], prompt_text_if_empty: str) -> None:
        if not self.is_mounted:
            return
        try:
            sel = self.query_one("#llm_config_model_select", Select)
            current_selected_model = sel.value
            sel.set_options(models)
            if models:
                sel.disabled = False
                sel.prompt = "Select Model"
                if current_selected_model in [m_val for _, m_val in models]:
                    sel.value = current_selected_model
                else:
                    sel.value = Select.BLANK
            else:
                sel.disabled = True
                sel.prompt = prompt_text_if_empty
                sel.value = Select.BLANK
                if self.model is not None:
                    self.model = None
        except NoMatches:
            logger.error("LLMConfigPanel: model select not found in update_models_list")

    def set_model_select_loading(
        self, loading: bool, loading_text: str = "Loading models..."
    ) -> None:
        if not self.is_mounted:
            return
        try:
            sel = self.query_one("#llm_config_model_select", Select)
            sel.disabled = loading
            if loading:
                sel.prompt = loading_text
            else:
                if not sel.disabled:
                    sel.prompt = "Select Model"
        except NoMatches:
            logger.warning("LLMConfigPanel: Model select not found in set_model_select_loading.")

    def watch_model(self, new_model: Optional[str]) -> None:
        if not self.is_mounted:
            return
        try:
            sel = self.query_one("#llm_config_model_select", Select)
            if sel.value != (new_model or Select.BLANK):
                sel.value = new_model or Select.BLANK
        except NoMatches:
            logger.warning("LLMConfigPanel: Model select not found in watch_model.")

    def get_llm_config(self) -> Dict[str, Any]:
        cfg = {
            "provider_hint": self.provider,
            "model_name": self.model,
            "temperature": self.temperature,
            "max_history_messages": self.max_history_messages,
        }
        return {k: v for k, v in cfg.items() if v is not None}



import logging
from typing import Any, Dict, Optional

from mongo_analyser.core.config_manager import (
    DEFAULT_SETTINGS,
    DEFAULT_THEME_NAME,
    VALID_THEMES,
    ConfigManager,
)
from mongo_analyser.dialogs import ConfirmDialog, ErrorDialog
from rich.text import Text
from textual import on
from textual.app import ComposeResult
from textual.containers import Container, VerticalScroll
from textual.css.query import NoMatches
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Button, Input, Label, Select, Static

logger = logging.getLogger(__name__)


class ConfigView(Container):
    config_save_feedback = reactive(Text(""))
    _feedback_timer_cv: Optional[Any] = None

    def __init__(
        self,
        *children: Widget,
        name: Optional[str] = None,
        id: Optional[str] = None,
        classes: Optional[str] = None,
        disabled: bool = False,
    ):
        super().__init__(*children, name=name, id=id, classes=classes, disabled=disabled)
        self._config_manager: Optional[ConfigManager] = None

    def on_mount(self) -> None:
        if hasattr(self.app, "config_manager"):
            self._config_manager = self.app.config_manager
            self.load_settings_to_ui()
        else:
            logger.error("ConfigManager not found on app. Cannot initialize ConfigView.")
            self.config_save_feedback = Text.from_markup(
                "[red]Error: ConfigManager not available.[/]"
            )

    def compose(self) -> ComposeResult:
        with VerticalScroll(id="config_view_scroll_panel"):
            yield Label("Theme:", classes="config_label")
            yield Select(
                [(name.replace("-", " ").title(), name) for name in VALID_THEMES],
                id="config_theme_select",
                value=DEFAULT_SETTINGS.get("theme", DEFAULT_THEME_NAME),
            )

            yield Label("Default Log Level:", classes="config_label")
            yield Select(
                [
                    ("DEBUG", "DEBUG"),
                    ("INFO", "INFO"),
                    ("WARNING", "WARNING"),
                    ("ERROR", "ERROR"),
                    ("CRITICAL", "CRITICAL"),
                    ("OFF", "OFF"),
                ],
                id="config_log_level_select",
                value=DEFAULT_SETTINGS.get("default_log_level", "INFO"),
            )

            yield Label("Schema Analysis - Default Sample Size:", classes="config_label")
            yield Input(
                id="config_schema_sample_size_input",
                value=str(DEFAULT_SETTINGS["schema_analysis_default_sample_size"]),
                placeholder="e.g., 1000",
                type="integer",
            )
            yield Label("Data Explorer - Default Sample Size:", classes="config_label")
            yield Input(
                id="config_explorer_sample_size_input",
                value=str(DEFAULT_SETTINGS["data_explorer_default_sample_size"]),
                placeholder="e.g., 10",
                type="integer",
            )

            yield Label("LLM Default Provider:", classes="config_label")
            yield Select(
                [("Ollama", "ollama"), ("OpenAI", "openai"), ("Google", "google")],
                id="config_llm_default_provider_select",
                value=DEFAULT_SETTINGS["llm_default_provider"],
                allow_blank=False,
            )
            yield Label("LLM Default Model - Ollama:", classes="config_label")
            yield Input(
                id="config_llm_model_ollama_input",
                value=DEFAULT_SETTINGS.get("llm_default_model_ollama", ""),
                placeholder="e.g., llama3:8b",
            )
            yield Label("LLM Default Model - OpenAI:", classes="config_label")
            yield Input(
                id="config_llm_model_openai_input",
                value=DEFAULT_SETTINGS.get("llm_default_model_openai", ""),
                placeholder="e.g., gpt-4o-mini",
            )
            yield Label("LLM Default Model - Google:", classes="config_label")
            yield Input(
                id="config_llm_model_google_input",
                value=DEFAULT_SETTINGS.get("llm_default_model_google", ""),
                placeholder="e.g., gemini-1.5-flash-latest",
            )

            yield Label("LLM Default Temperature:", classes="config_label")
            yield Input(
                id="config_llm_temperature_input",
                value=str(DEFAULT_SETTINGS["llm_default_temperature"]),
                placeholder="e.g., 0.7",
                type="number",
            )
            yield Label("LLM Default Max History Messages:", classes="config_label")
            yield Input(
                id="config_llm_max_history_input",
                value=str(DEFAULT_SETTINGS["llm_default_max_history"]),
                placeholder="e.g., 20 (0=all, -1=none)",
                type="integer",
            )

            yield Button("Save Configuration", variant="primary", id="config_save_button")
            yield Button("Reset to Defaults", variant="warning", id="config_reset_button")
            yield Static(self.config_save_feedback, id="config_save_feedback_label")

    def load_settings_to_ui(self) -> None:
        if not self._config_manager:
            return
        try:
            theme_select = self.query_one("#config_theme_select", Select)

            theme_to_set = self._config_manager.get_setting("theme", DEFAULT_THEME_NAME)

            current_options = theme_select._options
            if not current_options:
                theme_select.set_options(
                    [(name.replace("-", " ").title(), name) for name in VALID_THEMES]
                )
                current_options = theme_select._options
            if theme_to_set in [opt_val for _, opt_val in current_options]:
                theme_select.value = theme_to_set
            else:
                logger.warning(
                    f"Theme '{theme_to_set}' from config not in Select options. Defaulting to {DEFAULT_THEME_NAME}"
                )
                theme_select.value = DEFAULT_THEME_NAME

            self.query_one(
                "#config_log_level_select", Select
            ).value = self._config_manager.get_setting(
                "default_log_level", DEFAULT_SETTINGS.get("default_log_level", "INFO")
            )

            self.query_one("#config_schema_sample_size_input", Input).value = str(
                self._config_manager.get_setting(
                    "schema_analysis_default_sample_size",
                    DEFAULT_SETTINGS["schema_analysis_default_sample_size"],
                )
            )
            self.query_one("#config_explorer_sample_size_input", Input).value = str(
                self._config_manager.get_setting(
                    "data_explorer_default_sample_size",
                    DEFAULT_SETTINGS["data_explorer_default_sample_size"],
                )
            )
            self.query_one(
                "#config_llm_default_provider_select", Select
            ).value = self._config_manager.get_setting(
                "llm_default_provider", DEFAULT_SETTINGS["llm_default_provider"]
            )

            self.query_one("#config_llm_model_ollama_input", Input).value = (
                self._config_manager.get_setting("llm_default_model_ollama") or ""
            )
            self.query_one("#config_llm_model_openai_input", Input).value = (
                self._config_manager.get_setting("llm_default_model_openai") or ""
            )
            self.query_one("#config_llm_model_google_input", Input).value = (
                self._config_manager.get_setting("llm_default_model_google") or ""
            )
            self.query_one("#config_llm_temperature_input", Input).value = str(
                self._config_manager.get_setting(
                    "llm_default_temperature", DEFAULT_SETTINGS["llm_default_temperature"]
                )
            )
            self.query_one("#config_llm_max_history_input", Input).value = str(
                self._config_manager.get_setting(
                    "llm_default_max_history", DEFAULT_SETTINGS["llm_default_max_history"]
                )
            )

            logger.info("ConfigView: Settings loaded into UI.")
        except NoMatches as e:
            logger.error(f"ConfigView: Error finding widget to load settings: {e}")
        except Exception as e:
            logger.error(f"ConfigView: Unexpected error loading settings to UI: {e}", exc_info=True)

    def _collect_settings_from_ui(self) -> Dict[str, Any]:
        settings = {}
        try:
            theme_value = self.query_one("#config_theme_select", Select).value
            settings["theme"] = (
                str(theme_value)
                if theme_value != Select.BLANK and theme_value is not None
                else DEFAULT_THEME_NAME
            )
            if settings["theme"] not in VALID_THEMES:
                logger.warning(
                    f"Invalid theme '{settings['theme']}' collected from UI. Using default."
                )
                settings["theme"] = DEFAULT_THEME_NAME

            log_level_select_value = self.query_one("#config_log_level_select", Select).value
            settings["default_log_level"] = (
                str(log_level_select_value)
                if log_level_select_value != Select.BLANK
                else DEFAULT_SETTINGS.get("default_log_level", "INFO")
            )

            schema_sample_size_str = self.query_one("#config_schema_sample_size_input", Input).value
            try:
                settings["schema_analysis_default_sample_size"] = int(schema_sample_size_str)
            except ValueError:
                settings["schema_analysis_default_sample_size"] = DEFAULT_SETTINGS[
                    "schema_analysis_default_sample_size"
                ]
                logger.warning(
                    f"Invalid schema sample size '{schema_sample_size_str}', using default."
                )
            explorer_sample_size_str = self.query_one(
                "#config_explorer_sample_size_input", Input
            ).value
            try:
                settings["data_explorer_default_sample_size"] = int(explorer_sample_size_str)
            except ValueError:
                settings["data_explorer_default_sample_size"] = DEFAULT_SETTINGS[
                    "data_explorer_default_sample_size"
                ]
                logger.warning(
                    f"Invalid explorer sample size '{explorer_sample_size_str}', using default."
                )
            settings["llm_default_provider"] = str(
                self.query_one("#config_llm_default_provider_select", Select).value
            )
            ollama_model_val = self.query_one("#config_llm_model_ollama_input", Input).value.strip()
            settings["llm_default_model_ollama"] = ollama_model_val if ollama_model_val else None
            openai_model_val = self.query_one("#config_llm_model_openai_input", Input).value.strip()
            settings["llm_default_model_openai"] = openai_model_val if openai_model_val else None
            google_model_val = self.query_one("#config_llm_model_google_input", Input).value.strip()
            settings["llm_default_model_google"] = google_model_val if google_model_val else None
            llm_temp_str = self.query_one("#config_llm_temperature_input", Input).value
            try:
                settings["llm_default_temperature"] = float(llm_temp_str)
            except ValueError:
                settings["llm_default_temperature"] = DEFAULT_SETTINGS["llm_default_temperature"]
                logger.warning(f"Invalid LLM temperature '{llm_temp_str}', using default.")
            llm_hist_str = self.query_one("#config_llm_max_history_input", Input).value
            try:
                settings["llm_default_max_history"] = int(llm_hist_str)
            except ValueError:
                settings["llm_default_max_history"] = DEFAULT_SETTINGS["llm_default_max_history"]
                logger.warning(f"Invalid LLM max history '{llm_hist_str}', using default.")
        except NoMatches as e:
            logger.error(f"ConfigView: Error finding widget to collect settings: {e}")
            self.app.notify(
                f"UI Error: Could not read setting for a widget ({e.widget if e.widget else 'unknown'}).",
                severity="error",
                title="Config Error",
            )
        except Exception as e:
            logger.error(
                f"ConfigView: Unexpected error collecting settings from UI: {e}", exc_info=True
            )
            self.app.notify(
                "Unexpected error reading settings.", severity="error", title="Config Error"
            )
        return settings

    def _notify_restart_may_be_needed(self):
        self.app.notify(
            "Some settings (like default values or log level) will apply on next app start.",
            title="Restart Info",
            severity="information",
            timeout=7,
        )

    @on(Button.Pressed, "#config_save_button")
    async def save_configuration_button_pressed(self) -> None:
        if not self._config_manager:
            self.config_save_feedback = Text.from_markup(
                "[red]Error: ConfigManager not available for saving.[/]"
            )
            return
        settings_from_ui = self._collect_settings_from_ui()
        if not settings_from_ui:
            self.config_save_feedback = Text.from_markup(
                "[orange3]Warning: No valid settings collected from UI to save.[/]"
            )
            return

        self._config_manager.update_settings(settings_from_ui)
        if self._config_manager.save_config():
            self.config_save_feedback = Text.from_markup(
                "[green]Configuration saved successfully![/]"
            )
            self.app.notify("Configuration saved.", title="Save Success")
            if hasattr(self.app, "config_manager"):
                new_theme_name = self._config_manager.get_setting("theme", DEFAULT_THEME_NAME)
                if self.app.theme != new_theme_name:
                    self.app.theme = new_theme_name
                    logger.info(f"Theme changed to '{new_theme_name}' and applied from config.")
            self._notify_restart_may_be_needed()
        else:
            self.config_save_feedback = Text.from_markup(
                "[red]Error saving configuration. Check logs.[/]"
            )
            await self.app.push_screen(
                ErrorDialog("Save Error", "Could not save configuration. See logs for details.")
            )

    @on(Button.Pressed, "#config_reset_button")
    async def reset_configuration_button_pressed(self) -> None:
        if not self._config_manager:
            self.config_save_feedback = Text.from_markup(
                "[red]Error: ConfigManager not available for reset.[/]"
            )
            return
        confirm = await self.app.push_screen_wait(
            ConfirmDialog(
                "Reset Configuration",
                "Are you sure you want to reset all settings to their defaults?",
            )
        )
        if confirm:
            self._config_manager.update_settings(DEFAULT_SETTINGS.copy())
            self.load_settings_to_ui()
            if self._config_manager.save_config():
                self.config_save_feedback = Text.from_markup(
                    "[green]Configuration reset to defaults and saved.[/]"
                )
                self.app.notify("Configuration reset to defaults.", title="Reset Success")
                if hasattr(self.app, "config_manager"):
                    default_theme_name = self._config_manager.get_setting(
                        "theme", DEFAULT_THEME_NAME
                    )
                    if self.app.theme != default_theme_name:
                        self.app.theme = default_theme_name
                        logger.info(f"Theme reset to default '{default_theme_name}' and applied.")
                self._notify_restart_may_be_needed()
            else:
                self.config_save_feedback = Text.from_markup(
                    "[red]Error saving defaults after reset. Check logs.[/]"
                )

    def watch_config_save_feedback(self, new_feedback: Text) -> None:
        if self.is_mounted:
            try:
                lbl = self.query_one("#config_save_feedback_label", Static)
                lbl.update(new_feedback)
                if self._feedback_timer_cv is not None:
                    try:
                        self._feedback_timer_cv.stop()
                    except AttributeError:
                        if hasattr(self._feedback_timer_cv, "stop_no_wait"):
                            self._feedback_timer_cv.stop_no_wait()
                    self._feedback_timer_cv = None
                if new_feedback.plain:
                    self._feedback_timer_cv = self.set_timer(
                        4, lambda: setattr(self, "config_save_feedback", Text(""))
                    )
            except NoMatches:
                logger.warning(
                    "#config_save_feedback_label NOT FOUND in watch_config_save_feedback!"
                )
            except Exception as e:
                logger.error(
                    f"Error in watch_config_save_feedback (ConfigView): {e}", exc_info=True
                )

    def focus_default_widget(self) -> None:
        try:
            self.query_one("#config_theme_select", Select).focus()
        except NoMatches:
            logger.debug("ConfigView: Could not focus default select (#config_theme_select).")



from textual.app import ComposeResult
from textual.containers import Vertical
from textual.widgets import Markdown, Static


class ChatMessageWidget(Vertical):
    def __init__(self, role: str, content: str, **kwargs):
        super().__init__(**kwargs)
        self.role = role
        self.content = content

    def compose(self) -> ComposeResult:
        yield Static(self.role.upper(), classes=f"message-role role_{self.role.lower()}")
        yield Markdown(self.content, classes="message-content-box")



import logging
from typing import Optional, Tuple

from mongo_analyser.core.shared import redact_uri_password
from pymongo import MongoClient
from pymongo.database import Database as PyMongoDatabase
from pymongo.errors import ConfigurationError, ConnectionFailure, OperationFailure
from pymongo.server_api import ServerApi

logger = logging.getLogger(__name__)

_client: Optional[MongoClient] = None
_db: Optional[PyMongoDatabase] = None
_current_uri: Optional[str] = None
_current_db_name_arg: Optional[str] = None
_current_resolved_db_name: Optional[str] = None
_last_connection_error: Optional[Tuple[str, Optional[int]]] = None


def db_connection_active(
    uri: str,
    db_name: Optional[str] = None,
    server_timeout_ms: int = 5000,
    force_reconnect: bool = False,
    **kwargs,
) -> bool:
    global \
        _client, \
        _db, \
        _current_uri, \
        _current_db_name_arg, \
        _current_resolved_db_name, \
        _last_connection_error
    _last_connection_error = None

    redacted_uri_for_log = redact_uri_password(uri)

    if not force_reconnect and _client is not None and _db is not None and _current_uri == uri:
        target_db_name_for_check = db_name if db_name else _current_resolved_db_name

        if _db.name == target_db_name_for_check:
            try:
                _client.admin.command("ping")
                logger.debug(
                    f"Already connected to MongoDB (URI: {redacted_uri_for_log}, DB: {_db.name}). Ping OK."
                )
                return True
            except (ConnectionFailure, OperationFailure) as e:
                logger.warning(f"Existing MongoDB connection ping failed: {e}. Reconnecting.")
                _last_connection_error = (str(e), getattr(e, "code", None))
                _client, _db = None, None
        elif db_name and _db.name != db_name:
            logger.info(
                f"Switching DB context on existing client from '{_db.name}' to '{db_name}'."
            )
            try:
                _db = _client[db_name]
                _client.admin.command("ping")
                _current_db_name_arg = db_name
                _current_resolved_db_name = _db.name
                logger.info(f"Successfully switched DB context to '{_db.name}'.")
                return True
            except Exception as e:
                logger.error(f"Failed to switch DB context to '{db_name}' or ping failed: {e}")
                _last_connection_error = (str(e), getattr(e, "code", None))
                _client, _db = None, None
        else:
            _client, _db = None, None

    if _client is not None:
        _client.close()
        _client, _db = None, None

    _current_uri = None
    _current_db_name_arg = None
    _current_resolved_db_name = None

    try:
        logger.info(
            f"Attempting to connect to MongoDB: {redacted_uri_for_log}, TUI DB specified: {db_name}"
        )

        client_connect_options = {"serverSelectionTimeoutMS": server_timeout_ms, **kwargs}
        if uri.startswith("mongodb+srv://"):
            client_connect_options["server_api"] = ServerApi("1")
            logger.debug("SRV URI detected, applying ServerApi('1') option.")

        temp_client = MongoClient(uri, **client_connect_options)

        db_to_connect_to: Optional[str] = None

        if db_name:
            db_to_connect_to = db_name
            logger.info(
                f"Using database name explicitly provided via TUI/argument: '{db_to_connect_to}'"
            )
        else:
            try:
                db_from_uri_path = temp_client.get_database().name
                logger.info(f"Database name resolved from URI path: '{db_from_uri_path}'")
                db_to_connect_to = db_from_uri_path

                if db_from_uri_path.lower() in ("admin", "test", "config", "local"):
                    logger.warning(
                        f"URI resolved to default/system database '{db_from_uri_path}'. "
                        "If this is not your target data DB, please specify one in the TUI."
                    )
            except ConfigurationError:
                err_msg = (
                    f"MongoDB URI ('{redacted_uri_for_log}') does not specify a default database path, "
                    "and no database name was provided via the TUI. Cannot determine database context."
                )
                logger.error(err_msg)
                _last_connection_error = (err_msg, None)
                temp_client.close()

                return False

        if not db_to_connect_to:
            err_msg = (
                f"Fatal: Database name could not be determined for URI '{redacted_uri_for_log}'."
            )
            logger.error(err_msg)
            _last_connection_error = (err_msg, None)
            temp_client.close()
            return False

        temp_client.admin.command("ping")
        logger.debug("MongoDB server ping successful.")

        target_db_object = temp_client[db_to_connect_to]

        _client = temp_client
        _db = target_db_object
        _current_uri = uri
        _current_db_name_arg = db_name
        _current_resolved_db_name = _db.name

        logger.info(
            f"Successfully connected to MongoDB server. URI: '{redacted_uri_for_log}', "
            f"Effective DB Context: '{_current_resolved_db_name}'."
        )
        return True

    except ConfigurationError as e:
        logger.error(f"MongoDB client configuration error for URI '{redacted_uri_for_log}': {e}")
        _last_connection_error = (str(e), None)
        if _client:
            _client.close()
        _client, _db, _current_uri, _current_db_name_arg, _current_resolved_db_name = (
            None,
            None,
            None,
            None,
            None,
        )
        return False
    except (ConnectionFailure, OperationFailure) as e:
        logger.error(
            f"MongoDB connection/operation failure for URI '{redacted_uri_for_log}',"
            f" target DB '{db_name or 'from URI'}': {e}"
        )
        _last_connection_error = (str(e), getattr(e, "code", None))
        if _client:
            _client.close()
        _client, _db, _current_uri, _current_db_name_arg, _current_resolved_db_name = (
            None,
            None,
            None,
            None,
            None,
        )
        return False
    except Exception as e:
        logger.error(f"Unexpected error during MongoDB connection: {e}", exc_info=True)
        _last_connection_error = (str(e), None)
        if _client:
            _client.close()
        _client, _db, _current_uri, _current_db_name_arg, _current_resolved_db_name = (
            None,
            None,
            None,
            None,
            None,
        )
        return False


def get_last_connection_error_details() -> Optional[Tuple[str, Optional[int]]]:
    return _last_connection_error


def get_mongo_db() -> PyMongoDatabase:
    global _client, _db, _last_connection_error
    if _db is None or _client is None:
        raise ConnectionError(
            "Not connected to MongoDB or connection lost. Call db_connection_active first or reconnect."
        )
    try:
        _client.admin.command({"ping": 1})
    except (ConnectionFailure, OperationFailure) as e:
        logger.error(f"MongoDB connection lost when trying to get DB: {e}")
        _last_connection_error = (str(e), getattr(e, "code", None))
        disconnect_mongo()
        raise ConnectionError("MongoDB connection lost. Reconnect needed.") from e
    return _db


def get_mongo_client() -> Optional[MongoClient]:
    global _last_connection_error
    if _client:
        try:
            _client.admin.command("ping")
            return _client
        except (ConnectionFailure, OperationFailure) as e:
            logger.warning(f"Ping failed for existing client, considered disconnected: {e}")
            _last_connection_error = (str(e), getattr(e, "code", None))
            disconnect_mongo()
            return None
    return None


def get_current_uri() -> Optional[str]:
    return _current_uri


def get_current_resolved_db_name() -> Optional[str]:
    return _current_resolved_db_name


def disconnect_mongo() -> None:
    global \
        _client, \
        _db, \
        _current_uri, \
        _current_db_name_arg, \
        _current_resolved_db_name, \
        _last_connection_error
    if _client is not None:
        _client.close()
        logger.info("MongoDB client connection closed.")
    _client = None
    _db = None
    _current_uri = None
    _current_db_name_arg = None
    _current_resolved_db_name = None
    _last_connection_error = None


def disconnect_all_mongo() -> None:
    disconnect_mongo()



import json
import logging
import os
from pathlib import Path
from typing import Any, Dict, Optional

logger = logging.getLogger(__name__)

APP_DIR_NAME = "mongo_analyser"
DEFAULT_CONFIG_FILE_NAME = "config.json"

DEFAULT_THEME_NAME = "textual-dark"
VALID_THEMES = [
    "textual-dark",
    "textual-light",
    "nord",
    "gruvbox",
    "catppuccin-mocha",
    "dracula",
    "tokyo-night",
    "monokai",
    "flexoki",
    "catppuccin-latte",
    "solarized-light",
]

DEFAULT_SETTINGS = {
    "theme": DEFAULT_THEME_NAME,
    "default_log_level": "OFF",
    "schema_analysis_default_sample_size": 1000,
    "data_explorer_default_sample_size": 10,
    "llm_default_provider": "ollama",
    "llm_default_model_ollama": "gemma3:4b",
    "llm_default_model_openai": "gpt-4.1-nano",
    "llm_default_model_google": "models/gemini-2.0-flash-lite",
    "llm_default_temperature": 0.7,
    "llm_default_max_history": 20,
}


class ConfigManager:
    def __init__(self, base_app_data_dir_override: Optional[Path] = None):
        if base_app_data_dir_override:
            self._base_app_data_dir = base_app_data_dir_override.expanduser().resolve()
        else:
            self._base_app_data_dir = self._get_default_base_app_data_dir()

        try:
            logger.debug(
                f"ConfigManager: Ensuring base application data directory exists: {self._base_app_data_dir}"
            )
            self._base_app_data_dir.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            logger.error(
                f"CRITICAL FAILURE: Could not create or access base application data directory "
                f"{self._base_app_data_dir}: {e}. Configs, logs, and exports will likely fail.",
                exc_info=True,
            )

        self._config_file_path = self._base_app_data_dir / DEFAULT_CONFIG_FILE_NAME
        self._config: Dict[str, Any] = {}
        logger.debug(
            f"ConfigManager initialized. Base data dir: {self._base_app_data_dir}, Config file: {self._config_file_path}"
        )
        self.load_config()

    def _get_default_base_app_data_dir(self) -> Path:
        xdg_data_home = os.environ.get("XDG_DATA_HOME")
        if xdg_data_home:
            base_path = Path(xdg_data_home) / APP_DIR_NAME
        else:
            base_path = Path.home() / ".local" / "share" / APP_DIR_NAME
        return base_path

    def get_base_app_data_dir(self) -> Path:
        """Returns the root directory used by the application for its data."""
        return self._base_app_data_dir

    def get_config_file_path(self) -> Path:
        """Returns the full path to the config.json file."""
        return self._config_file_path

    def _get_or_create_subdir(self, subdir_name: str) -> Path:
        """Helper to get a subdirectory path within the base app data dir, creating it if necessary."""
        subdir = self._base_app_data_dir / subdir_name
        try:
            subdir.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            logger.error(
                f"Could not create subdirectory {subdir}: {e}. Operations in this directory might fail."
            )

        return subdir

    def get_logs_dir(self) -> Path:
        return self._get_or_create_subdir("logs")

    def get_chats_dir(self) -> Path:
        return self._get_or_create_subdir("chats")

    def get_exports_dir(self) -> Path:
        return self._get_or_create_subdir("exports")

    def load_config(self) -> None:
        loaded_settings = {}
        if self._config_file_path.exists() and self._config_file_path.is_file():
            try:
                with open(self._config_file_path, "r", encoding="utf-8") as f:
                    loaded_settings = json.load(f)
                logger.info(f"Configuration loaded from {self._config_file_path}")
            except (IOError, json.JSONDecodeError) as e:
                logger.error(
                    f"Error loading configuration from {self._config_file_path}: {e}. Using defaults."
                )
        else:
            logger.info(
                f"Configuration file not found at {self._config_file_path}. Using defaults and will create on save."
            )

        self._config = DEFAULT_SETTINGS.copy()
        self._config.update(loaded_settings)

        if self._config.get("theme") not in VALID_THEMES:
            logger.warning(
                f"Loaded theme '{self._config.get('theme')}' is invalid. Resetting to default."
            )
            self._config["theme"] = DEFAULT_SETTINGS["theme"]

        loaded_log_level = self._config.get("default_log_level")
        valid_log_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "OFF"]
        if not (isinstance(loaded_log_level, str) and loaded_log_level.upper() in valid_log_levels):
            logger.warning(
                f"Loaded default_log_level '{loaded_log_level}' is invalid. Resetting to default."
            )
            self._config["default_log_level"] = DEFAULT_SETTINGS["default_log_level"]
        else:
            self._config["default_log_level"] = loaded_log_level.upper()

    def save_config(self) -> bool:
        try:
            self._base_app_data_dir.mkdir(parents=True, exist_ok=True)

            logger.debug(f"Attempting to save configuration to: {self._config_file_path}")
            with open(self._config_file_path, "w", encoding="utf-8") as f:
                json.dump(self._config, f, indent=2)
            logger.info(f"Configuration saved successfully to {self._config_file_path}")
            return True
        except OSError as e:
            logger.error(
                f"OSError saving configuration to {self._config_file_path}. "
                f"Error: {e.strerror} (errno {e.errno}). Path involved: {e.filename}",
                exc_info=True,
            )
            return False
        except Exception as e:
            logger.error(
                f"Unexpected error saving configuration to {self._config_file_path}: {e}",
                exc_info=True,
            )
            return False

    def get_setting(self, key: str, default: Optional[Any] = None) -> Any:
        if key == "theme":
            theme_value = self._config.get(key, DEFAULT_SETTINGS.get(key))
            return theme_value if theme_value in VALID_THEMES else DEFAULT_THEME_NAME

        if key in self._config:
            return self._config[key]

        return DEFAULT_SETTINGS.get(key, default)

    def update_setting(self, key: str, value: Any) -> None:
        if key == "theme":
            if value not in VALID_THEMES:
                logger.warning(
                    f"Attempted to set invalid theme '{value}'. Using default '{DEFAULT_THEME_NAME}'."
                )
                self._config[key] = DEFAULT_THEME_NAME
            else:
                self._config[key] = value
            return

        if key == "default_log_level":
            valid_log_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "OFF"]
            if isinstance(value, str) and value.upper() in valid_log_levels:
                self._config[key] = value.upper()
            else:
                logger.warning(
                    f"Attempted to set invalid default_log_level '{value}'. Using default '{DEFAULT_SETTINGS['default_log_level']}'."
                )
                self._config[key] = DEFAULT_SETTINGS["default_log_level"]
            return

        self._config[key] = value

    def get_all_settings(self) -> Dict[str, Any]:
        effective_settings = DEFAULT_SETTINGS.copy()

        effective_settings.update(self._config)

        if effective_settings.get("theme") not in VALID_THEMES:
            effective_settings["theme"] = DEFAULT_THEME_NAME

        loaded_log_level = effective_settings.get("default_log_level")
        valid_log_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "OFF"]
        if not (isinstance(loaded_log_level, str) and loaded_log_level.upper() in valid_log_levels):
            effective_settings["default_log_level"] = DEFAULT_SETTINGS["default_log_level"]
        else:
            effective_settings["default_log_level"] = loaded_log_level.upper()

        return effective_settings

    def update_settings(self, new_settings: Dict[str, Any]) -> None:
        for key, value in new_settings.items():
            self.update_setting(key, value)



from . import config_manager, db, shared
from .analyser import SchemaAnalyser
from .extractor import DataExtractor

__all__ = [
    "DataExtractor",
    "SchemaAnalyser",
    "config_manager",
    "db",
    "shared",
]



from .base import LLMChat
from .google import GoogleChat
from .ollama import OllamaChat
from .openai import OpenAIChat

__all__ = [
    "GoogleChat",
    "LLMChat",
    "OllamaChat",
    "OpenAIChat",
]



import gzip
import json
import logging
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pytz
from bson import Binary, Decimal128, Int64, ObjectId
from bson.binary import UuidRepresentation
from pymongo import DESCENDING
from pymongo.errors import (
    ConnectionFailure as PyMongoConnectionFailure,
)
from pymongo.errors import (
    OperationFailure as PyMongoOperationFailure,
)
from textual.worker import (
    Worker,
    WorkerCancelled,
    get_current_worker,
)

from . import db as db_manager
from . import shared

logger = logging.getLogger(__name__)

_BSON_UUID_SUBTYPE_STANDARD = 4
_BSON_UUID_SUBTYPE_LEGACY_PYTHON = 3


class DataExtractor:
    @staticmethod
    def _infer_type_val(value: Any) -> str:
        if isinstance(value, bool):
            return "bool"
        elif isinstance(value, Int64):
            return "int64"
        elif isinstance(value, int):
            return "int32"
        elif isinstance(value, float):
            return "double"
        elif isinstance(value, str):
            return "str"
        elif isinstance(value, datetime):
            return "datetime"
        elif isinstance(value, ObjectId):
            return "ObjectId"
        elif isinstance(value, uuid.UUID):
            return "UUID"
        elif isinstance(value, Binary):
            if value.subtype == _BSON_UUID_SUBTYPE_LEGACY_PYTHON:
                return "binary<UUID (legacy)>"
            if value.subtype == _BSON_UUID_SUBTYPE_STANDARD:
                return "binary<UUID>"
            return shared.binary_type_map.get(value.subtype, f"binary<subtype {value.subtype}>")
        elif isinstance(value, list):
            if not value:
                return "array<empty>"

            element_types = set()
            has_actual_value = False
            for item in value:
                inferred_item_type = DataExtractor._infer_type_val(item)
                element_types.add(inferred_item_type)
                if inferred_item_type != "null":
                    has_actual_value = True

            if not has_actual_value:
                return "array<null>"

            distinct_types_no_null = {t for t in element_types if t != "null"}

            if len(distinct_types_no_null) == 1:
                return f"array<{list(distinct_types_no_null)[0]}>"
            elif not distinct_types_no_null and "null" in element_types:
                return "array<null>"
            else:
                return "array<mixed>"
        elif isinstance(value, dict):
            return "dict"
        elif isinstance(value, Decimal128):
            return "decimal128"
        elif value is None:
            return "null"
        else:
            return f"unknown<{type(value).__name__}>"

    @staticmethod
    def _convert_single_value(
        val: Any,
        schema_type_str: Optional[str],
        tz: Union[pytz.timezone, None],
        items_schema_for_array_elements: Optional[Dict] = None,
    ) -> Any:
        if val is None:
            return None

        type_to_check = schema_type_str or DataExtractor._infer_type_val(val)

        if isinstance(val, list):
            if (
                type_to_check == "array<dict>"
                and items_schema_for_array_elements
                and isinstance(items_schema_for_array_elements, dict)
            ):
                return [
                    DataExtractor.convert_to_json_compatible(
                        item, items_schema_for_array_elements, tz
                    )
                    if isinstance(item, dict)
                    else DataExtractor._convert_single_value(item, None, tz)
                    for item in val
                ]
            else:
                item_type_str_for_elements = None
                if (
                    type_to_check
                    and type_to_check.startswith("array<")
                    and type_to_check.endswith(">")
                ):
                    item_type_str_for_elements = type_to_check[len("array<") : -1]

                return [
                    DataExtractor._convert_single_value(item, item_type_str_for_elements, tz)
                    for item in val
                ]

        if isinstance(val, uuid.UUID):
            return str(val)

        if isinstance(val, Binary) and val.subtype in (
            _BSON_UUID_SUBTYPE_STANDARD,
            _BSON_UUID_SUBTYPE_LEGACY_PYTHON,
        ):
            try:
                py_uuid = None
                if val.subtype == _BSON_UUID_SUBTYPE_STANDARD:
                    py_uuid = val.as_uuid(UuidRepresentation.STANDARD)
                elif val.subtype == _BSON_UUID_SUBTYPE_LEGACY_PYTHON:
                    try:
                        py_uuid = val.as_uuid(UuidRepresentation.PYTHON_LEGACY)
                    except ValueError:
                        py_uuid = val.as_uuid()

                if py_uuid:
                    return str(py_uuid)
                else:
                    logger.warning(
                        f"Binary UUID subtype {val.subtype} could not be converted to Python UUID, falling to hex."
                    )
                    return val.hex()
            except Exception as e:
                logger.warning(
                    f"Could not convert Binary UUID subtype {val.subtype} (Exception: {e}), falling back to hex."
                )
                return val.hex()

        if type_to_check == "binary<UUID>" or type_to_check == "binary<UUID (legacy)>":
            if isinstance(val, str) and len(val) == 36:
                try:
                    uuid.UUID(val)
                    return val
                except ValueError:
                    pass
            if isinstance(val, Binary):
                return val.hex()
            return str(val)

        if (
            type_to_check == "binary<ObjectId>"
            or type_to_check == "ObjectId"
            or isinstance(val, ObjectId)
        ):
            return str(val)

        if type_to_check == "datetime" or isinstance(val, datetime):
            dt_to_convert = val
            if val.tzinfo is None or val.tzinfo.utcoffset(val) is None:
                dt_to_convert = pytz.utc.localize(val)

            if tz:
                return dt_to_convert.astimezone(tz).isoformat()
            return dt_to_convert.isoformat()

        if type_to_check == "str":
            return str(val)
        if type_to_check in ("int32", "int64") or isinstance(val, (int, Int64)):
            return int(val)
        if type_to_check == "bool" or isinstance(val, bool):
            return bool(val)
        if type_to_check == "double" or isinstance(val, float):
            return float(val)
        if type_to_check == "decimal128" or isinstance(val, Decimal128):
            return str(val.to_decimal())

        if isinstance(val, Binary):
            return val.hex()

        if isinstance(val, (str, int, float, bool, dict)):
            return val

        logger.warning(
            f"Value {str(val)[:50]} of type {type(val)} with schema type hint '{schema_type_str}' fell through to string conversion."
        )
        return str(val)

    @staticmethod
    def convert_to_json_compatible(
        document: Dict, schema_for_current_level: Dict, tz: Union[pytz.timezone, None]
    ) -> Dict:
        processed_document: Dict = {}
        for key, value in document.items():
            if value is None:
                processed_document[key] = None
                continue

            field_schema_definition = schema_for_current_level.get(key)
            type_str_from_schema: Optional[str] = None
            items_sub_schema: Optional[Dict] = None

            if isinstance(field_schema_definition, dict):
                type_str_from_schema = field_schema_definition.get("type")

                if (
                    type_str_from_schema
                    and type_str_from_schema.startswith("array<")
                    and isinstance(field_schema_definition.get("items"), dict)
                ):
                    if type_str_from_schema == "array<dict>":
                        items_sub_schema = field_schema_definition.get("items")

                elif not type_str_from_schema and isinstance(value, dict):
                    processed_document[key] = DataExtractor.convert_to_json_compatible(
                        value,
                        field_schema_definition,
                        tz,
                    )
                    continue

            processed_document[key] = DataExtractor._convert_single_value(
                value, type_str_from_schema, tz, items_sub_schema
            )
        return processed_document

    @staticmethod
    def extract_data(
        mongo_uri: str,
        db_name: str,
        collection_name: str,
        schema: Dict,
        output_file: Union[str, Path],
        tz: Union[None, pytz.timezone],
        batch_size: int,
        limit: int,
        server_timeout_ms: int = 5000,
    ) -> None:
        output_p = Path(output_file)
        output_p.parent.mkdir(parents=True, exist_ok=True)

        worker_instance: Optional[Worker] = None
        try:
            worker_instance = get_current_worker()
        except Exception:
            pass

        if worker_instance and worker_instance.is_cancelled:
            logger.info("Data extraction cancelled before database connection.")
            raise WorkerCancelled()

        if not db_manager.db_connection_active(
            uri=mongo_uri, db_name=db_name, server_timeout_ms=server_timeout_ms
        ):
            if worker_instance and worker_instance.is_cancelled:
                logger.info("Data extraction cancelled during/after database connection attempt.")
                raise WorkerCancelled()
            raise PyMongoConnectionFailure(
                f"MongoDB connection failed for data extraction: {mongo_uri}, DB: {db_name}"
            )

        database = db_manager.get_mongo_db()
        collection = database[collection_name]
        data_cursor = None

        try:
            if worker_instance and worker_instance.is_cancelled:
                logger.info("Data extraction cancelled before finding documents.")
                raise WorkerCancelled()

            data_cursor = (
                collection.find(no_cursor_timeout=True)
                .sort("_id", DESCENDING)
                .batch_size(batch_size)
            )
            if limit >= 0:
                data_cursor = data_cursor.limit(limit)
                logger.info(
                    f"Reading up to {limit} newest records from {db_name}.{collection_name}..."
                )
            else:
                logger.info(
                    f"Reading all records (newest first) from {db_name}.{collection_name}..."
                )

            count = 0
            with gzip.open(output_p, "wt", encoding="utf-8") as f:
                f.write("[\n")
                first_doc = True
                for doc in data_cursor:
                    if worker_instance and worker_instance.is_cancelled:
                        logger.info(f"Data extraction cancelled by worker at document {count + 1}.")
                        if not first_doc:
                            f.write("\n")
                        raise WorkerCancelled()

                    count += 1
                    converted_doc = DataExtractor.convert_to_json_compatible(doc, schema, tz)

                    if not first_doc:
                        f.write(",\n")

                    json.dump(converted_doc, f, indent=None)
                    first_doc = False

                    if count % batch_size == 0:
                        logger.info(f"Processed {count} documents...")
                        if worker_instance:
                            worker_instance.update_message(f"Processed {count} documents...")

                if not first_doc:
                    f.write("\n")
                f.write("]\n")
                logger.info(f"Successfully extracted {count} documents to {output_p}")

        except WorkerCancelled:
            logger.warning(
                f"Extraction process cancelled. Output file {output_p} may be incomplete or not a valid JSON array."
            )
            raise
        except PyMongoOperationFailure as e:
            if worker_instance and worker_instance.is_cancelled:
                logger.warning(f"Extraction cancelled during MongoDB operation: {e}")
                raise WorkerCancelled()
            logger.error(
                f"MongoDB operation failure during data extraction from {db_name}.{collection_name}: {e}"
            )
            raise
        except IOError as e:
            if worker_instance and worker_instance.is_cancelled:
                logger.warning(f"Extraction cancelled during file IO: {e}")
                raise WorkerCancelled()
            logger.error(f"Failed to write to output file {output_p}: {e}")
            raise
        finally:
            if data_cursor is not None:
                data_cursor.close()
                logger.debug("MongoDB cursor closed for data extraction.")

    @staticmethod
    def get_newest_documents(
        mongo_uri: str,
        db_name: str,
        collection_name: str,
        sample_size: int,
        server_timeout_ms: int = 5000,
    ) -> List[Dict]:
        if not db_manager.db_connection_active(
            uri=mongo_uri, db_name=db_name, server_timeout_ms=server_timeout_ms
        ):
            raise PyMongoConnectionFailure(
                f"MongoDB connection failed for sampling ({db_name}.{collection_name})"
            )

        database = db_manager.get_mongo_db()
        collection = database[collection_name]
        query_cursor = None

        try:
            if sample_size <= 0:
                logger.warning(
                    "Sample size must be positive for fetching newest documents. Returning empty list."
                )
                return []

            projection_doc: Optional[Dict[str, int]] = None

            query_cursor = (
                collection.find(projection=projection_doc)
                .sort("_id", DESCENDING)
                .limit(sample_size)
            )

            raw_documents = list(query_cursor)
            processed_docs: List[Dict] = []
            for doc in raw_documents:
                processed_doc = {}
                for key, value in doc.items():
                    if isinstance(value, ObjectId):
                        processed_doc[key] = str(value)
                    elif isinstance(value, datetime):
                        processed_doc[key] = value.isoformat()
                    elif isinstance(value, uuid.UUID):
                        processed_doc[key] = str(value)
                    elif isinstance(value, Binary):
                        if value.subtype in (
                            _BSON_UUID_SUBTYPE_STANDARD,
                            _BSON_UUID_SUBTYPE_LEGACY_PYTHON,
                        ):
                            try:
                                representation = (
                                    UuidRepresentation.STANDARD
                                    if value.subtype == _BSON_UUID_SUBTYPE_STANDARD
                                    else UuidRepresentation.PYTHON_LEGACY
                                )
                                processed_doc[key] = str(value.as_uuid(representation))
                            except Exception:
                                processed_doc[key] = value.hex()[:64] + (
                                    "..." if len(value.hex()) > 64 else ""
                                )
                        else:
                            hex_val = value.hex()
                            processed_doc[key] = (
                                f"binary_hex:{hex_val[:64]}{'...' if len(hex_val) > 64 else ''}"
                            )
                    elif isinstance(value, Decimal128):
                        processed_doc[key] = str(value.to_decimal())
                    elif isinstance(value, (list, dict, str, int, float, bool)) or value is None:
                        try:
                            val_str_for_size_check = (
                                json.dumps(value, default=str)
                                if isinstance(value, (dict, list))
                                else str(value)
                            )
                            if len(val_str_for_size_check) > 500:
                                processed_doc[key] = (
                                    f"{type(value).__name__}(too large to display inline)"
                                )
                            else:
                                processed_doc[key] = value
                        except TypeError:
                            processed_doc[key] = (
                                f"unserializable_type:{type(value).__name__}:{str(value)[:50]}"
                            )

                    else:
                        processed_doc[key] = (
                            f"unhandled_type:{type(value).__name__}:{str(value)[:50]}"
                        )
                processed_docs.append(processed_doc)

            logger.info(
                f"Fetched {len(processed_docs)} newest documents from {db_name}.{collection_name}"
            )
            return processed_docs

        except PyMongoOperationFailure as e:
            logger.error(
                f"MongoDB operation failed during document fetching ({db_name}.{collection_name}): {e}"
            )
            raise
        except Exception as e:
            logger.error(
                f"Unexpected error during document fetching ({db_name}.{collection_name}): {e}",
                exc_info=True,
            )
            raise
        finally:
            if query_cursor is not None:
                query_cursor.close()
                logger.debug("MongoDB cursor closed for get_newest_documents.")



from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Label


class ConfirmDialog(ModalScreen[bool]):
    BINDINGS = [Binding("escape", "dismiss(False)", show=False)]

    def __init__(self, title: str, message: str, yes_label: str = "Yes", no_label: str = "No"):
        super().__init__()
        self._title = title
        self._message = message
        self._yes_label = yes_label
        self._no_label = no_label

    def compose(self) -> ComposeResult:
        with Vertical() as v_layout:
            v_layout.border_title = self._title
            yield Label(self._message)
            with Horizontal():
                yield Button(self._yes_label, variant="primary", id="yes_button")
                yield Button(self._no_label, id="no_button")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes_button":
            self.dismiss(True)
        elif event.button.id == "no_button":
            self.dismiss(False)



from textual.app import ComposeResult
from textual.css.query import NoMatches
from textual.widgets import Input, Static


class ChatInput(Static):
    def compose(self) -> ComposeResult:
        yield Input(placeholder="Type your message here...", id="chat_internal_input")

    @property
    def value(self) -> str:
        try:
            return self.query_one(Input).value
        except NoMatches:
            return ""

    @value.setter
    def value(self, new_value: str) -> None:
        try:
            self.query_one(Input).value = new_value
        except NoMatches:
            pass

    def clear(self) -> None:
        self.value = ""

    def focus(self, scroll_visible: bool = True) -> None:
        try:
            self.query_one(Input).focus(scroll_visible)
        except NoMatches:
            pass



import logging
import uuid
from collections import Counter
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

from bson import Binary, Decimal128, Int64, ObjectId
from pymongo.errors import (
    ConnectionFailure as PyMongoConnectionFailure,
)
from pymongo.errors import (
    OperationFailure as PyMongoOperationFailure,
)
from pymongo.synchronous.collection import Collection as PyMongoCollection

from . import db as db_manager
from . import shared

logger = logging.getLogger(__name__)


class SchemaAnalyser:
    @staticmethod
    def _make_hashable(item: Any) -> Any:
        if isinstance(item, dict):
            return frozenset((k, SchemaAnalyser._make_hashable(v)) for k, v in sorted(item.items()))
        elif isinstance(item, list):
            return tuple(SchemaAnalyser._make_hashable(i) for i in item)
        else:
            return item

    @staticmethod
    def extract_schema_and_stats(
        document: Dict,
        schema: Optional[Dict] = None,
        stats: Optional[Dict] = None,
        prefix: str = "",
    ) -> Tuple[Dict, Dict]:
        if schema is None:
            schema = {}
        if stats is None:
            stats = {}

        for key, value in document.items():
            full_key = f"{prefix}.{key}" if prefix else key

            if full_key not in stats:
                stats[full_key] = {
                    "values": set(),
                    "count": 0,
                    "type_counts": Counter(),
                    "numeric_min": float("inf"),
                    "numeric_max": float("-inf"),
                    "date_min": None,
                    "date_max": None,
                    "value_frequencies": Counter(),
                    "array_element_stats": {
                        "values": set(),
                        "type_counts": Counter(),
                        "numeric_min": float("inf"),
                        "numeric_max": float("-inf"),
                        "date_min": None,
                        "date_max": None,
                        "value_frequencies": Counter(),
                    },
                }

            stats[full_key]["count"] += 1

            if isinstance(value, dict):
                SchemaAnalyser.extract_schema_and_stats(value, schema, stats, full_key)
            elif isinstance(value, list):
                SchemaAnalyser.handle_array(value, schema, stats, full_key)
            else:
                SchemaAnalyser.handle_simple_value(
                    value, schema, stats[full_key], full_key, is_array_element=False
                )

        return schema, stats

    @staticmethod
    def handle_array(
        value: List,
        schema: Dict,
        stats: Dict,
        full_key: str,
    ) -> None:
        current_field_stats = stats[full_key]

        if not value:
            schema[full_key] = {"type": "array<empty>"}
        else:
            element_types_for_schema = Counter()
            sample_elements_for_schema = value[:10]

            for elem in sample_elements_for_schema:
                if isinstance(elem, dict):
                    element_types_for_schema.update(["dict"])
                elif isinstance(elem, ObjectId):
                    element_types_for_schema.update(["ObjectId"])
                elif isinstance(elem, uuid.UUID):
                    element_types_for_schema.update(["UUID"])
                elif isinstance(elem, Binary):
                    subtype_str = shared.binary_type_map.get(
                        elem.subtype, f"binary<subtype {elem.subtype}>"
                    )
                    element_types_for_schema.update([subtype_str])
                elif isinstance(elem, bool):
                    element_types_for_schema.update(["bool"])
                elif isinstance(elem, Int64):
                    element_types_for_schema.update(["int64"])
                elif isinstance(elem, int):
                    element_types_for_schema.update(["int32"])
                elif isinstance(elem, float):
                    element_types_for_schema.update(["double"])
                elif isinstance(elem, Decimal128):
                    element_types_for_schema.update(["decimal128"])
                elif isinstance(elem, datetime):
                    element_types_for_schema.update(["datetime"])
                else:
                    element_types_for_schema.update([type(elem).__name__])

            if len(element_types_for_schema) == 1:
                dominant_type = element_types_for_schema.most_common(1)[0][0]
                schema[full_key] = {"type": f"array<{dominant_type}>"}
            elif element_types_for_schema:
                schema[full_key] = {"type": "array<mixed>"}
            else:
                schema[full_key] = {"type": "array<unknown>"}

            for elem in value:
                SchemaAnalyser.handle_simple_value(
                    elem,
                    {},
                    current_field_stats["array_element_stats"],
                    full_key,
                    is_array_element=True,
                )

        try:
            hashable_value = SchemaAnalyser._make_hashable(value)
            current_field_stats["values"].add(hashable_value)
        except TypeError:
            current_field_stats["values"].add(f"unhashable_array_len_{len(value)}")

    @staticmethod
    def handle_simple_value(
        value: Any,
        schema: Dict,
        stats_dict_to_update: Dict,
        full_key: str,
        is_array_element: bool,
    ) -> None:
        stats_dict_to_update.setdefault("values", set())
        stats_dict_to_update.setdefault("type_counts", Counter())
        stats_dict_to_update.setdefault("numeric_min", float("inf"))
        stats_dict_to_update.setdefault("numeric_max", float("-inf"))
        stats_dict_to_update.setdefault("date_min", None)
        stats_dict_to_update.setdefault("date_max", None)
        stats_dict_to_update.setdefault("value_frequencies", Counter())

        if isinstance(value, ObjectId):
            value_type_name = "binary<ObjectId>"
        elif isinstance(value, uuid.UUID):
            value_type_name = "binary<UUID>"
        elif isinstance(value, Binary):
            value_type_name = shared.binary_type_map.get(
                value.subtype, f"binary<subtype {value.subtype}>"
            )
        elif isinstance(value, bool):
            value_type_name = "bool"
        elif isinstance(value, Int64):
            value_type_name = "int64"
        elif isinstance(value, int):
            value_type_name = "int32"
        elif isinstance(value, float):
            value_type_name = "double"
        elif isinstance(value, Decimal128):
            value_type_name = "decimal128"
        elif isinstance(value, datetime):
            value_type_name = "datetime"
        else:
            value_type_name = type(value).__name__

        if not is_array_element:
            schema[full_key] = {"type": value_type_name}
            try:
                stats_dict_to_update["values"].add(SchemaAnalyser._make_hashable(value))
            except:
                stats_dict_to_update["values"].add(f"unhashable_value_type_{type(value).__name__}")

        stats_dict_to_update["type_counts"].update([value_type_name])

        if isinstance(value, (int, float, Int64, Decimal128)):
            num_val = float(value.to_decimal()) if isinstance(value, Decimal128) else float(value)
            stats_dict_to_update["numeric_min"] = min(
                stats_dict_to_update.get("numeric_min", float("inf")), num_val
            )
            stats_dict_to_update["numeric_max"] = max(
                stats_dict_to_update.get("numeric_max", float("-inf")), num_val
            )
        elif isinstance(value, str):
            if len(value) < 256:
                stats_dict_to_update["value_frequencies"].update([value])
        elif isinstance(value, datetime):
            cur_min = stats_dict_to_update.get("date_min")
            cur_max = stats_dict_to_update.get("date_max")
            if cur_min is None or value < cur_min:
                stats_dict_to_update["date_min"] = value
            if cur_max is None or value > cur_max:
                stats_dict_to_update["date_max"] = value

    @staticmethod
    def get_collection(
        uri: str, db_name: str, collection_name: str, server_timeout_ms: int = 5000
    ) -> PyMongoCollection:
        if not db_manager.db_connection_active(
            uri=uri, db_name=db_name, server_timeout_ms=server_timeout_ms
        ):
            raise PyMongoConnectionFailure(
                f"Failed to establish or verify active connection to MongoDB for {db_name}"
            )
        database = db_manager.get_mongo_db()
        return database[collection_name]

    @staticmethod
    def list_collection_names(uri: str, db_name: str, server_timeout_ms: int = 5000) -> List[str]:
        if not db_manager.db_connection_active(
            uri=uri, db_name=db_name, server_timeout_ms=server_timeout_ms
        ):
            raise PyMongoConnectionFailure(
                f"Failed to establish or verify active connection to MongoDB for listing collections in {db_name}"
            )

        database = db_manager.get_mongo_db()
        try:
            return sorted(database.list_collection_names())
        except PyMongoOperationFailure as e:
            logger.error(f"MongoDB operation failure listing collections for DB '{db_name}': {e}")
            raise

    @staticmethod
    def infer_schema_and_field_stats(
        collection: PyMongoCollection, sample_size: int, batch_size: int = 1000
    ) -> Tuple[Dict, Dict]:
        schema: Dict = {}
        stats: Dict = {}
        total_docs = 0

        try:
            docs = (
                collection.find().batch_size(batch_size)
                if sample_size < 0
                else collection.aggregate([{"$sample": {"size": sample_size}}]).batch_size(
                    batch_size
                )
            )
            for doc in docs:
                total_docs += 1
                schema, stats = SchemaAnalyser.extract_schema_and_stats(doc, schema, stats)
                if 0 < sample_size <= total_docs:
                    break
        except PyMongoOperationFailure as e:
            logger.error(
                f"Error during schema inference on {collection.database.name}.{collection.name}: {e}"
            )
            raise

        final_stats_summary: Dict = {}
        for key, field_stat_data in stats.items():
            values_set = field_stat_data.get("values", set())
            count = field_stat_data.get("count", 0)
            type_counts = field_stat_data.get("type_counts", Counter())
            val_freqs = field_stat_data.get("value_frequencies", Counter())
            arr_stats = field_stat_data.get("array_element_stats", {})
            arr_type_counts = arr_stats.get("type_counts", Counter())
            arr_val_freqs = arr_stats.get("value_frequencies", Counter())

            cardinality = len(values_set)
            missing_count = total_docs - count
            missing_pct = (missing_count / total_docs * 100) if total_docs else 0

            processed: Dict[str, Any] = {
                "cardinality": cardinality,
                "missing_percentage": missing_pct,
                "type_distribution": dict(type_counts.most_common(5)),
            }

            num_min = field_stat_data.get("numeric_min", float("inf"))
            if num_min != float("inf"):
                processed["numeric_min"] = num_min
                processed["numeric_max"] = field_stat_data.get("numeric_max", num_min)

            dmin = field_stat_data.get("date_min")
            if dmin is not None:
                processed["date_min"] = dmin.isoformat()
                processed["date_max"] = field_stat_data.get("date_max", dmin).isoformat()

            if val_freqs:
                processed["top_values"] = dict(val_freqs.most_common(5))

            if arr_type_counts:
                arr_processed: Dict[str, Any] = {
                    "type_distribution": dict(arr_type_counts.most_common(5))
                }
                arr_num_min = arr_stats.get("numeric_min", float("inf"))
                if arr_num_min != float("inf"):
                    arr_processed["numeric_min"] = arr_num_min
                    arr_processed["numeric_max"] = arr_stats.get("numeric_max", arr_num_min)
                arr_dmin = arr_stats.get("date_min")
                if arr_dmin is not None:
                    arr_processed["date_min"] = arr_dmin.isoformat()
                    arr_processed["date_max"] = arr_stats.get("date_max", arr_dmin).isoformat()
                if arr_val_freqs:
                    arr_processed["top_values"] = dict(arr_val_freqs.most_common(5))
                processed["array_elements"] = arr_processed

            final_stats_summary[key] = processed

        sorted_schema = dict(sorted(schema.items()))
        sorted_stats = dict(sorted(final_stats_summary.items()))
        return sorted_schema, sorted_stats

    @staticmethod
    def schema_to_hierarchical(schema: Dict) -> Dict:
        hierarchical: Dict = {}
        for field, details in schema.items():
            parts = field.split(".")
            cur = hierarchical
            for p in parts[:-1]:
                cur = cur.setdefault(p, {})
            cur[parts[-1]] = {"type": details.get("type")}
        return hierarchical



from .chat_input import ChatInput
from .chat_message_list import ChatMessageList
from .chat_message_widget import ChatMessageWidget
from .collection_stats_widget import CollectionStatsWidget
from .llm_config_panel import LLMConfigPanel

__all__ = [
    "ChatInput",
    "ChatMessageList",
    "ChatMessageWidget",
    "CollectionStatsWidget",
    "LLMConfigPanel",
]



import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterator, List, Optional

logger = logging.getLogger(__name__)


class LLMChat(ABC):
    def __init__(self, model_name: str, **kwargs: Any):
        self.model_name = model_name
        self.client_config = kwargs
        logger.info(f"Initializing {self.__class__.__name__} with model: {self.model_name}")
        self.client = self._initialize_client(**self.client_config)

    @abstractmethod
    def _initialize_client(self, **kwargs: Any) -> Any:
        pass

    @abstractmethod
    def send_message(self, message: str, history: List[Dict[str, str]] = None) -> str:
        pass

    @abstractmethod
    def stream_message(self, message: str, history: List[Dict[str, str]] = None) -> Iterator[str]:
        pass

    @staticmethod
    @abstractmethod
    def list_models(client_config: Optional[Dict[str, Any]] = None) -> List[str]:
        pass

    def get_available_models(self) -> List[str]:
        return self.__class__.list_models(client_config=self.client_config)

    def format_history(self, history: List[Dict[str, str]] = None) -> List[Dict[str, str]]:
        formatted_history = []
        if history:
            for msg in history:
                role = msg.get("role")
                content = msg.get("content")
                if role and content:
                    if role == "ai":
                        role = "assistant"
                    formatted_history.append({"role": role, "content": content})
                else:
                    logger.warning(f"Skipping malformed message in history: {msg}")
        return formatted_history



from .chat_view import ChatView
from .config_view import ConfigView
from .data_explorer_view import DataExplorerView
from .db_connection_view import DBConnectionView
from .schema_analysis_view import SchemaAnalysisView

__all__ = [
    "ChatView",
    "ConfigView",
    "DBConnectionView",
    "DataExplorerView",
    "SchemaAnalysisView",
]



import logging
from typing import Union

from urllib.parse import quote_plus, urlparse, urlunparse

logger = logging.getLogger(__name__)

binary_type_map = {
    0: "binary<generic>",
    1: "binary<function>",
    3: "binary<UUID (legacy)>",
    4: "binary<UUID>",
    5: "binary<MD5>",
}


def build_mongo_uri(
    host: str,
    port: Union[str, int],
    username: Union[str, None] = None,
    password: Union[str, None] = None,
    params: Union[str, None] = None,
) -> str:
    base_uri = "mongodb://"
    if username and password:
        base_uri += f"{quote_plus(username)}:{quote_plus(password)}@"
    elif username:
        base_uri += f"{quote_plus(username)}@"
    base_uri += f"{host}:{port}/"
    if params:
        base_uri += f"?{params}"
    return base_uri


def redact_uri_password(uri: str) -> str:
    try:
        parsed = urlparse(uri)
        if parsed.password:
            netloc_parts = parsed.netloc.split("@", 1)
            auth_part_present = len(netloc_parts) > 1 and ":" in netloc_parts[0]

            if auth_part_present:
                user_pass_part = netloc_parts[0]
                host_part = netloc_parts[1]
                username = user_pass_part.split(":", 1)[0]
                new_netloc = f"{username}:********@{host_part}"
                parsed = parsed._replace(netloc=new_netloc)
                return urlunparse(parsed)
        return uri
    except Exception as e:
        logger.error(f"Error redacting URI password: {e}")
        return uri



from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Label


class ErrorDialog(ModalScreen[None]):
    BINDINGS = [Binding("escape", "dismiss", show=False)]

    def __init__(self, title: str, message: str):
        super().__init__()
        self._title = title
        self._message = message

    def compose(self) -> ComposeResult:
        with Vertical() as v_layout:
            v_layout.border_title = self._title

            yield Label(self._message, classes="dialog_message_error")
            with Center():
                yield Button("OK", variant="error", id="ok_button")

    def on_mount(self) -> None:
        self.query_one("#ok_button", Button).focus()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "ok_button":
            self.dismiss()



import logging
import os
import re
from typing import Any, Dict, Iterator, List, Optional

import litellm

from .interface import LLMChat

logger = logging.getLogger(__name__)

MODEL_BLOCKLISTS = {
    "openai": [
        r"babbage.*",
        r"davinci.*",
        r"curie.*",
        r"ada.*",
        r"dall-e.*",
        r"text-embedding.*",
        r"*-embedding-.*",
        r"tts-.*",
        r"whisper-.*",
        r"gpt-3\.5-turbo-instruct.*",
        r"text-moderation-.*",
    ],
    "google": [
        r"models/embedding-.*",
        r"models/aqa",
        r".*-tts",
        r".*-transcribe",
        r".*-vision",
    ],
    "google_vertex": [
        r"text-bison.*",
        r"chat-bison.*",
        r"embedding-gecko.*",
    ],
    "ollama": [
        r".*embed.*",
        r"nomic-embed-text",
        r"mxbai-embed-large",
        r"snowflake-arctic-embed",
        r"all-minilm",
        r"bge-.*",
        r"distiluse-base-multilingual-cased",
        r"e5-.*",
        r"gemma:.*-text",
        r"sentence-transformers/.*",
        r"vision",
        r"llava",
    ],
    "common_suffixes_to_avoid": [
        "-instruct",
        "-code",
        "-edit",
        "-vision",
        "-embed",
        "-tts",
        "-stt",
        "-preview",
        "-experimental",
    ],
}


def _is_model_blocked(model_name: str, provider: Optional[str]) -> bool:
    """Checks if a model name matches any regex pattern in the blocklists."""
    block_patterns: List[str] = []
    if provider:
        block_patterns.extend(MODEL_BLOCKLISTS.get(provider.lower(), []))

    for pattern in block_patterns:
        try:
            if re.search(pattern, model_name, re.IGNORECASE):
                logger.debug(
                    f"Model '{model_name}' blocked by pattern '{pattern}' for provider '{provider}'."
                )
                return True
        except re.error as e:
            logger.warning(f"Regex error in blocklist pattern '{pattern}': {e}")
    return False


class LiteLLMChat(LLMChat):
    def __init__(self, model_name: str, provider_hint: Optional[str] = None, **kwargs: Any):
        self.raw_model_name = model_name
        self.provider_hint = (
            provider_hint.lower() if provider_hint else self._guess_provider(model_name)
        )

        self.config_params = kwargs

        fq_model_name = self.raw_model_name
        if self.provider_hint == "ollama" and not self.raw_model_name.startswith("ollama/"):
            fq_model_name = f"ollama/{self.raw_model_name}"
        elif (
            self.provider_hint == "google"
            and not self.raw_model_name.startswith("gemini/")
            and "/" not in self.raw_model_name
        ):
            fq_model_name = f"gemini/{self.raw_model_name}"

        super().__init__(fq_model_name, **kwargs)

    def _guess_provider(self, model_name: str) -> Optional[str]:
        model_lower = model_name.lower()
        if model_lower.startswith("gpt-") or "openai/" in model_lower or "gpt-4" in model_lower:
            return "openai"
        if "gemini" in model_lower or "google/" in model_lower or model_lower.startswith("models/"):
            return "google"
        if model_lower.startswith("ollama/"):
            return "ollama"
        if "claude" in model_lower or "anthropic/" in model_lower:
            return "anthropic"
        if "mistral" in model_lower and not model_lower.startswith("ollama/"):
            if (
                "mistral/" in model_lower
                or "open-mistral" in model_lower
                or "mixtral" in model_lower
            ):
                return "mistral"
        if "azure/" in model_lower:
            return "azure"

        return None

    def _initialize_client(self, **kwargs: Any) -> Any:
        self.api_key = self.config_params.get("api_key")
        self.base_url = self.config_params.get("base_url")
        self.temperature = float(self.config_params.get("temperature", 0.7))
        self.max_tokens = int(self.config_params.get("max_tokens", 2048))
        self.system_prompt = self.config_params.get("system_prompt")

        if self.provider_hint == "openai" and self.api_key and not os.getenv("OPENAI_API_KEY"):
            os.environ["OPENAI_API_KEY"] = self.api_key
            logger.debug("Set OPENAI_API_KEY from config for this session.")
        if self.provider_hint == "google" and self.api_key and not os.getenv("GOOGLE_API_KEY"):
            os.environ["GOOGLE_API_KEY"] = self.api_key
            logger.debug("Set GOOGLE_API_KEY from config for this session.")

        logger.info(
            f"LiteLLMChat configured for effective model '{self.model_name}' "
            f"(Provider hint: {self.provider_hint}, Raw name from panel: {self.raw_model_name}, "
            f"Temp: {self.temperature}, MaxOutputTokens: {self.max_tokens})"
        )
        return None

    def _prepare_messages_payload(
        self,
        message: str,
        history: List[Dict[str, str]] = None,
    ) -> List[Dict[str, str]]:
        messages_payload: List[Dict[str, str]] = []
        if self.system_prompt:
            messages_payload.append({"role": "system", "content": self.system_prompt})

        if history:
            messages_payload.extend(history)

        messages_payload.append({"role": "user", "content": message})
        return messages_payload

    def send_message(self, message: str, history: List[Dict[str, str]] = None) -> str:
        messages_payload = self._prepare_messages_payload(message, history)

        call_kwargs: Dict[str, Any] = {
            "model": self.model_name,
            "messages": messages_payload,
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
        }
        if self.api_key:
            call_kwargs["api_key"] = self.api_key
        if self.base_url:
            call_kwargs["api_base"] = self.base_url

        for k, v in self.config_params.items():
            if (
                k
                not in [
                    "model_name",
                    "provider_hint",
                    "api_key",
                    "base_url",
                    "temperature",
                    "max_tokens",
                    "system_prompt",
                    "max_history_messages",
                ]
                and k not in call_kwargs
            ):
                call_kwargs[k] = v

        try:
            logged_kwargs = {
                k: (v if k != "messages" else f"<{len(v)} messages>")
                for k, v in call_kwargs.items()
            }
            logger.debug(
                f"Calling LiteLLM completion for {self.model_name} with kwargs: {logged_kwargs}"
            )

            response = litellm.completion(**call_kwargs)

            assistant_response = ""
            if (
                response.choices
                and response.choices[0].message
                and response.choices[0].message.content
            ):
                assistant_response = response.choices[0].message.content
            return assistant_response.strip()
        except Exception as e:
            logger.error(
                f"LiteLLM completion error for model {self.model_name}: {e}", exc_info=True
            )

            return (
                f"Error from LLM ({self.raw_model_name}): {e.__class__.__name__} - {str(e)[:100]}"
            )

    def stream_message(self, message: str, history: List[Dict[str, str]] = None) -> Iterator[str]:
        logger.warning(
            "stream_message is called but not fully utilized by ChatView's current send mechanism."
        )
        messages_payload = self._prepare_messages_payload(message, history)
        call_kwargs: Dict[str, Any] = {
            "model": self.model_name,
            "messages": messages_payload,
            "stream": True,
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
        }
        if self.api_key:
            call_kwargs["api_key"] = self.api_key
        if self.base_url:
            call_kwargs["api_base"] = self.base_url

        try:
            response_stream = litellm.completion(**call_kwargs)
            for chunk in response_stream:
                if chunk.choices and chunk.choices[0].delta and chunk.choices[0].delta.content:
                    yield chunk.choices[0].delta.content
        except Exception as e:
            logger.error(f"LiteLLM streaming error for model {self.model_name}: {e}", exc_info=True)
            yield f"Error streaming from LLM ({self.raw_model_name}): {e.__class__.__name__} - {str(e)[:100]}"

    @staticmethod
    def list_models(
        provider: Optional[str] = None, client_config: Optional[Dict[str, Any]] = None
    ) -> List[str]:
        logger.info(f"LiteLLMChat.list_models called for provider: {provider}")
        cfg = client_config or {}

        final_model_list: List[str] = []

        if provider and provider.lower() == "ollama":
            logger.info("Attempting to list Ollama models dynamically.")
            ollama_base_url = (
                cfg.get("base_url")
                or os.getenv("OLLAMA_HOST")
                or os.getenv("LITELLM_OLLAMA_BASE_URL")
                or "http://localhost:11434"
            )
            ollama_timeout = cfg.get("timeout", 10)

            try:
                import ollama as ollama_client_lib

                client = ollama_client_lib.Client(host=ollama_base_url, timeout=ollama_timeout)
                models_info = client.list()

                if models_info and "models" in models_info:
                    raw_ollama_models = [
                        m.get("model", m.get("name"))
                        for m in models_info["models"]
                        if m.get("model", m.get("name"))
                    ]
                    raw_ollama_models = sorted(list(set(raw_ollama_models)))
                    logger.info(
                        f"Dynamically listed {len(raw_ollama_models)} unique models from Ollama host {ollama_base_url} before filtering."
                    )

                    for model_name_fq in raw_ollama_models:
                        base_model_name = model_name_fq.split(":")[0]
                        if not _is_model_blocked(
                            base_model_name, "ollama"
                        ) and not _is_model_blocked(model_name_fq, "ollama"):
                            final_model_list.append(base_model_name)
                        else:
                            logger.debug(
                                f"Ollama dynamic model '{model_name_fq}' (base: {base_model_name}) filtered out by blocklist."
                            )
                    final_model_list = sorted(list(set(final_model_list)))
                else:
                    logger.warning(
                        f"Dynamic Ollama listing from {ollama_base_url} returned no 'models' or unexpected format."
                    )
            except ImportError:
                logger.warning(
                    "Python 'ollama' package not installed. Cannot dynamically list Ollama models. Install with: pip install ollama"
                )
            except Exception as e:
                logger.error(
                    f"Failed to dynamically list Ollama models from {ollama_base_url}: {e}. Check Ollama server.",
                    exc_info=False,
                )

            if not final_model_list:
                logger.info(
                    "Falling back to static list for Ollama models from litellm.model_list."
                )
                for m_fq in litellm.model_list:
                    if m_fq.startswith("ollama/"):
                        model_name_part = m_fq.split("/", 1)[1]
                        if not _is_model_blocked(model_name_part, "ollama"):
                            final_model_list.append(model_name_part)
                final_model_list = sorted(list(set(final_model_list)))
            logger.info(f"Prepared {len(final_model_list)} Ollama models for TUI.")
            return final_model_list

        elif provider:
            logger.info(f"Listing models for non-Ollama provider: {provider}")

            provider_env_key_map = {
                "openai": "OPENAI_API_KEY",
                "google": "GOOGLE_API_KEY",
                "anthropic": "ANTHROPIC_API_KEY",
            }
            api_key_from_config = cfg.get("api_key")
            original_env_val = None
            env_key_to_set = provider_env_key_map.get(provider.lower())

            if env_key_to_set and api_key_from_config and not os.getenv(env_key_to_set):
                original_env_val = os.getenv(env_key_to_set)
                os.environ[env_key_to_set] = api_key_from_config
                logger.debug(f"Temporarily set {env_key_to_set} for model listing.")

            try:
                for model_id_fq in litellm.model_list:
                    model_provider_guess = ""
                    model_base_name = model_id_fq

                    if "/" in model_id_fq:
                        parts = model_id_fq.split("/", 1)
                        model_provider_guess = parts[0].lower()
                        model_base_name = parts[1]
                    else:
                        if model_id_fq.startswith("gpt-"):
                            model_provider_guess = "openai"
                        elif "gemini" in model_id_fq:
                            model_provider_guess = "google"

                    if (
                        model_provider_guess == provider.lower()
                        or (
                            (provider.lower() == "openai" and model_provider_guess == "")
                            and model_id_fq.startswith("gpt-")
                        )
                        or (
                            (provider.lower() == "google" and model_provider_guess == "")
                            and "gemini" in model_id_fq
                        )
                    ):
                        if not _is_model_blocked(
                            model_base_name, provider
                        ) and not _is_model_blocked(model_id_fq, provider):
                            final_model_list.append(model_base_name)
                        else:
                            logger.debug(
                                f"Model '{model_id_fq}' (base: {model_base_name}) for provider '{provider}' filtered out by blocklist."
                            )
            finally:
                if (
                    env_key_to_set
                    and api_key_from_config
                    and os.getenv(env_key_to_set) == api_key_from_config
                ):
                    if original_env_val is None:
                        del os.environ[env_key_to_set]
                        logger.debug(f"Cleared temporarily set {env_key_to_set}.")
                    else:
                        os.environ[env_key_to_set] = original_env_val
                        logger.debug(f"Restored {env_key_to_set}.")

            final_model_list = sorted(list(set(final_model_list)))
            logger.info(
                f"Prepared {len(final_model_list)} models for provider '{provider}' for TUI."
            )
            return final_model_list

        else:
            logger.warning(
                "LiteLLMChat.list_models called without a specific provider (and not Ollama). Returning empty list."
            )
            return []
